earth_radius = 6371008.8

factors = {
    "centimeters": earth_radius * 100,
    "centimetres": earth_radius * 100,
    "degrees": earth_radius / 111325,
    "feet": earth_radius * 3.28084,
    "inches": earth_radius * 39.370,
    "kilometers": earth_radius / 1000,
    "kilometres": earth_radius / 1000,
    "meters": earth_radius,
    "metres": earth_radius,
    "miles": earth_radius / 1609.344,
    "millimeters": earth_radius * 1000,
    "millimetres": earth_radius * 1000,
    "nautical_miles": earth_radius / 1852,
    "radians": 1,
    "yards": earth_radius / 1.0936,
}

units_factors = {
    "centimeters": 100,
    "centimetres": 100,
    "degrees": 1 / 111325,
    "feet": 3.28084,
    "inches": 39.370,
    "kilometers": 1 / 1000,
    "kilometres": 1 / 1000,
    "meters": 1,
    "metres": 1,
    "miles": 1 / 1609.344,
    "millimeters": 1000,
    "millimetres": 1000,
    "nautical_miles": 1 / 1852,
    "radians": 1 / earth_radius,
    "yards": 1 / 1.0936,
}

area_factors = {
    "acres": 0.000247105,
    "centimeters": 10000,
    "centimetres": 10000,
    "feet": 10.763910417,
    "inches": 1550.003100006,
    "kilometers": 0.000001,
    "kilometres": 0.000001,
    "meters": 1,
    "metres": 1,
    "miles": 3.86e-7,
    "millimeters": 1000000,
    "millimetres": 1000000,
    "yards": 1.195990046,
}

geometry_types = {
    "Point": "Point",
    "LineString": "LineString",
    "Polygon": "Polygon",
    "MultiPoint": "MultiPoint",
    "MultiLineString": "MultiLineString",
    "MultiPolygon": "MultiPolygon",
}



from typing import Dict, List, Union

from turf.helpers import FeatureCollection
from turf.rectangle_grid import rectangle_grid


def square_grid(
    bbox: List[float],
    n_cells: Union[int, float],
    options: Dict = {},
) -> FeatureCollection:
    """
    Creates a square of rectangles from a bounding box, Feature or FeatureCollection.

    :param bbox: Array extent in [minX, minY, maxX, maxY] order
    :param n_cells: number of each cell, in units
    :param options: Optional parameters
        [options["units"]]: units ("degrees", "radians", "miles", "kilometers")
                            of the given cell_width and cell_height
        [options["mask"]]: if passed a Polygon or MultiPolygon here,
                           the grid Points will be created only inside it
        [options["properties"]]: passed to each point of the grid

    :returns: FeatureCollection of a grid of polygons
    """

    return rectangle_grid(bbox, n_cells, n_cells, options)



from typing import Dict, List, Sequence, TypeVar, Union
from decimal import Decimal, getcontext
from math import sqrt

from turf.helpers import Feature, LineString, Point
from turf.helpers import feature_collection, line_string, multi_line_string
from turf.invariant import get_coords_from_features


ctx = getcontext()
ctx.prec = 28

PointFeature = TypeVar("PointFeature", Dict, Feature, Point)
LineFeature = TypeVar("LineFeature", Dict, Feature, LineString)


def boolean_point_on_line(
    point: PointFeature, line: LineFeature, options: Dict = {}
) -> bool:
    """
    Returns True if a point is on a line else False.
    Accepts a optional parameter to ignore the start and end vertices of the linestring.

    :param point: {Point} GeoJSON Point
    :param line: {LineString} GeoJSON LineString
    :param options: Optional parameters
        [options["ignoreEndVertices"]=False] whether to ignore the start and end vertices

    :return: boolean True/False if point is on line
    """
    if not isinstance(options, dict):
        options = {}

    point_on_line = False

    ignore_end_vertices = options.get("ignoreEndVertices", False)

    point_coord = get_coords_from_features(point, ("Point",))
    line_coords = get_coords_from_features(line, ("LineString",))

    for i in range(len(line_coords) - 1):
        if ignore_end_vertices:
            # ignore if point_coord are the line start
            if (i == 0) and (point_coord == line_coords[i]):
                continue

            # ignore if point_coord are the line end
            if ((i + 1) == (len(line_coords) - 1)) and (
                point_coord == line_coords[i + 1]
            ):
                continue

        point_on_line = point_on_segment(
            point_coord, line_coords[i], line_coords[i + 1]
        )

        if point_on_line:
            break

    return point_on_line


def point_on_segment(
    point: List, segment_start: List, segment_end: List, epsilon: float = 1.0e-14
) -> bool:
    """
    Checks if a given point is on a line or not.

    Since this is a comparison of floats, I use the Decimal module of python


    :param point: Coordinates of a point
    :param segment_start: Coordinates of the start line
    :param segment_end: Coordinates of the line end
    :return: bool
    """

    len_segment = Decimal(
        sqrt(
            pow(segment_end[0] - segment_start[0], 2)
            + pow(segment_end[1] - segment_start[1], 2)
        )
    )

    len_point_seg_1 = Decimal(
        sqrt(pow(point[0] - segment_start[0], 2) + pow(point[1] - segment_start[1], 2))
    )

    len_point_seg_2 = Decimal(
        sqrt(pow(segment_end[0] - point[0], 2) + pow(segment_end[1] - point[1], 2))
    )

    return abs(len_segment - len_point_seg_1 - len_point_seg_2) <= epsilon



from abc import ABC
from typing import List, Dict, Any, Iterable, Union, Sequence

from turf.utils.error_codes import error_code_messages
from turf.utils.exceptions import InvalidInput
from turf.utils.helpers import get_input_dimensions


all_geometry_types = [
    "Point",
    "LineString",
    "Polygon",
    "MultiPoint",
    "MultiLineString",
    "MultiPolygon",
]


class Geometry(ABC):
    """
    Base class for Point, LineString and Polygon sub-classes.
    """

    def __init__(self, coordinates: Iterable, geometry_type) -> None:
        self._check_input(coordinates)
        self.coordinates = coordinates
        self.type = geometry_type

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.coordinates})"

    def __eq__(self, other: Any) -> bool:
        try:
            equality = self.type == other.type
            equality &= self.coordinates == other.coordinates
        except AttributeError:
            return False

        return equality

    def get(self, attribute: str, default=None) -> Any:
        try:
            return getattr(self, attribute)
        except AttributeError:
            return default

    @classmethod
    def from_geojson(cls, geojson: Dict) -> Any:
        try:
            coords = geojson["coordinates"]
        except KeyError:
            raise InvalidInput(error_code_messages["InvalidCoordinates"])

        return geometry(cls.__name__, coords, as_geojson=False)

    def to_geojson(self) -> Dict:
        """
        Translates the object into a GeoJSON feature.

        :return: a GeoJSON feature as a dict
        """

        return {"type": self.type, "coordinates": self.coordinates}


class Point(Geometry):
    """
    Class for creating Point objects with certain coordinates.
    Equivalent to a GeoJSON Point.
    """

    def __init__(self, coordinates: Sequence) -> None:
        Geometry.__init__(self, coordinates, "Point")

    @staticmethod
    def _check_input(coordinates: Sequence) -> None:
        """
        Checks input given to Point class, and raises an error if input is invalid.

        :param coordinates: input coordinates
        """

        if (
            get_input_dimensions(coordinates) == 1
            and len(coordinates) >= 2
            and all(isinstance(x, (int, float)) for x in coordinates)
        ):
            return

        raise InvalidInput(error_code_messages["InvalidPointInput"])


class MultiPoint(Point):
    def __init__(self, coordinates: Sequence) -> None:
        Point.__init__(self, coordinates)
        self.type = "MultiPoint"

    def _check_input(self, coordinates: Sequence) -> None:
        """
        Checks input given to MultiPoint class, and raises an error if input is invalid.

        :param coordinates: input coordinates
        """

        if get_input_dimensions(coordinates) != 2:
            raise InvalidInput(error_code_messages["InvalidMultiInput"] + "of Points")

        for coord in coordinates:
            super(MultiPoint, self)._check_input(coord)


class LineString(Geometry):
    """
    Class for creating LineString objects with certain coordinates.
    Equivalent to a GeoJSON LineString.
    """

    def __init__(self, coordinates: Sequence) -> None:
        Geometry.__init__(self, coordinates, geometry_type="LineString")

    def _check_input(self, coordinates: Sequence) -> None:
        """
        Checks input given to LineString class, and raises an error if input is invalid.

        :param coordinates: input coordinates
        """

        if get_input_dimensions(coordinates) != 2:
            raise InvalidInput(error_code_messages["InvalidLineStringInput"])

        if len(coordinates) >= 2 and all(
            all(isinstance(x, (int, float)) for x in y) for y in coordinates
        ):
            return

        raise InvalidInput(error_code_messages["InvalidLinePoints"])


class MultiLineString(LineString):
    def __init__(self, coordinates: Sequence) -> None:
        LineString.__init__(self, coordinates)
        self.type = "MultiLineString"

    def _check_input(self, coordinates: Sequence) -> None:
        """
        Checks input given to MultiLineString class, and raises an error if input is invalid.

        :param coordinates: input coordinates
        """

        if get_input_dimensions(coordinates) != 3:
            raise InvalidInput(
                error_code_messages["InvalidMultiInput"] + "of LineStrings"
            )

        for coord in coordinates:
            super(MultiLineString, self)._check_input(coord)


class Polygon(Geometry):
    """
    Class for creating Polygon objects with certain coordinates.
    Equivalent to a GeoJSON Polygon.
    """

    def __init__(self, coordinates: Sequence) -> None:
        Geometry.__init__(self, coordinates, geometry_type="Polygon")

    def _check_input(self, coordinates: Sequence) -> None:
        """
        Checks input given to Polygon class, and raises an error if input is invalid.

        :param coordinates: input coordinates
        """

        if get_input_dimensions(coordinates) != 3:
            raise InvalidInput(error_code_messages["InvalidPolygonInput"])

        for ring in coordinates:
            if not isinstance(ring, list):
                raise InvalidInput(error_code_messages["InvalidLinearRing"])

            if len(ring) < 4:
                raise InvalidInput(error_code_messages["InvalidLinearRing"])

            if ring[-1] != ring[0]:
                raise InvalidInput(error_code_messages["InvalidFirstLastPoints"])


class MultiPolygon(Polygon):
    def __init__(self, coordinates: Sequence) -> None:
        Polygon.__init__(self, coordinates)
        self.type = "MultiPolygon"

    def _check_input(self, coordinates: Sequence) -> None:
        """
        Checks input given to MultiPolygon class, and raises an error if input is invalid.

        :param coordinates: input coordinates
        """

        if get_input_dimensions(coordinates) != 4:
            raise InvalidInput(error_code_messages["InvalidMultiInput"] + "of Polygons")

        for coord in coordinates:
            super(MultiPolygon, self)._check_input(coord)


class FeatureType:
    """
    Parent class for Feature and FeatureCollection.
    """

    def __init__(self, feature_type: str) -> None:
        self.type = feature_type

    def get(self, attribute: str, default=None) -> Any:
        try:
            return getattr(self, attribute)
        except AttributeError:
            return default


class Feature(FeatureType):
    """
    Class that encapsulates a certain geometry, along with its properties.
    Equivalent to a GeoJSON feature.
    """

    def __init__(
        self,
        geom: Union[
            Dict,
            Point,
            LineString,
            Polygon,
            MultiPoint,
            MultiLineString,
            MultiPolygon,
        ],
        properties: Union[Dict, None] = None,
    ) -> None:
        geom = self._check_input(geom)

        FeatureType.__init__(self, feature_type="Feature")

        self.geometry = geom
        self.properties = properties or {}

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.geometry})"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Feature):
            return False

        equality = self.geometry.type == other.geometry.type
        equality &= self.geometry.coordinates == other.geometry.coordinates

        return equality

    @staticmethod
    def _check_input(
        geom: Union[
            Dict,
            Point,
            LineString,
            Polygon,
            MultiPoint,
            MultiLineString,
            MultiPolygon,
        ]
    ) -> Union[Point, LineString, Polygon, MultiPoint, MultiLineString, MultiPolygon,]:
        """
        Checks input given to Feature class, and converts to object if input is in dict form.

        :param geom: input geometry
        :return: geometry object
        """

        if not any(isinstance(geom, eval(cls)) for cls in all_geometry_types):
            if isinstance(geom, dict):
                feat_type = geom.get("type", "nonexistent")
                try:
                    return eval(feat_type).from_geojson(geom)
                except (NameError, AttributeError):
                    raise InvalidInput(
                        error_code_messages["InvalidGeometry"](all_geometry_types)
                    )
            else:
                raise InvalidInput(
                    error_code_messages["InvalidGeometry"](all_geometry_types)
                )

        return geom

    def to_geojson(self) -> Dict:
        """
        Translates the object into a GeoJSON feature.

        :return: a GeoJSON feature as a dict
        """

        geojson = {
            "type": "Feature",
            "properties": self.properties,
            "geometry": self.geometry.to_geojson(),
        }

        if self.get("bbox"):
            geojson["bbox"] = self.get("bbox")

        return geojson


class FeatureCollection(FeatureType):
    """
    Class that encapsulates a group of features in a FeatureCollection.
    Equivalent to a GeoJSON FeatureCollection.
    """

    def __init__(self, features: Sequence = None) -> None:
        features = self._check_input(features)

        FeatureType.__init__(self, feature_type="FeatureCollection")

        self.features = features or []

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({[feat.get('geometry').get('type') for feat in self.features]})"

    @staticmethod
    def _check_input(
        features: Sequence,
    ) -> List[Union[Dict, Feature]]:
        """
        Checks input given to FeatureCollection class, and converts to list of
        Feature objects if input is in dict form.

        :param features: input features
        :return: a list of the feature objects
        """

        if not isinstance(features, list):
            raise InvalidInput(error_code_messages["InvalidFeatureCollection"])

        eval_feats = []
        for feat in features:
            if not isinstance(feat, Feature):
                if isinstance(feat, dict):
                    feat_type = feat.get("geometry", {}).get("type", "nonexistent")
                    try:
                        geom = eval(feat_type).from_geojson(feat.get("geometry", {}))
                        properties = feat.get("properties", None)
                        feat_from_geojson = feature(
                            geom, properties=properties, as_geojson=False
                        )

                        eval_feats.append(feat_from_geojson)
                    except NameError:
                        raise InvalidInput(
                            error_code_messages["InvalidGeometry"](all_geometry_types)
                        )
                else:
                    raise InvalidInput(
                        error_code_messages["InvalidGeometry"](all_geometry_types)
                    )

            else:
                eval_feats.append(feat)

        return eval_feats

    def to_geojson(self) -> Dict:
        """
        Translates the object into a GeoJSON feature.

        :return: a GeoJSON feature as a dict
        """

        geojson = {"type": "FeatureCollection", "features": []}

        for f in self.features:
            geojson["features"].append(f.to_geojson())

        return geojson


def feature(
    geom: (Dict, Point, LineString, Polygon, MultiPoint, MultiLineString, MultiPolygon),
    properties: Dict = None,
    options: Dict = None,
    as_geojson: bool = True,
) -> Union[Feature, Dict]:
    """
    Wraps a GeoJSON Geometry in a GeoJSON Feature.

    :param geom: input geometry
    :param properties: a dictionary of key-value pairs to add as properties
    :param options: an options dictionary:
        [options["bbox"] Bounding Box Array [west, south, east, north] associated with the Feature
        [options["id"] Identifier associated with the Feature
    :param as_geojson: whether the return value should be a geojson
    :return: a GeoJSON feature
    """

    if not options:
        options = {}

    if not properties:
        properties = {}

    feat = Feature(geom, properties)

    if "id" in options:
        feat.id = options["id"]

    if "bbox" in options:
        feat.bbox = options["bbox"]

    return feat.to_geojson() if as_geojson else feat


def feature_collection(
    features: Sequence, options: Dict = None, as_geojson: bool = True
) -> Union[FeatureCollection, Dict]:
    """
    Takes one or more Feature and creates a FeatureCollection.

    :param features: input features
    :param options: an options dictionary:
        [options["bbox"] Bounding Box Array [west, south, east, north] associated with the Feature
        [options["id"] Identifier associated with the Feature
    :param as_geojson: whether the return value should be a geojson
    :return: a FeatureCollection of Features
    """

    if not options:
        options = {}

    feat_collection = FeatureCollection(features)

    if "id" in options:
        feat_collection.id = options["id"]

    if "bbox" in options:
        feat_collection.bbox = options["bbox"]

    return feat_collection.to_geojson() if as_geojson else feat_collection


def geometry(
    geom_type: str, coordinates: Sequence, as_geojson: bool = True
) -> Union[
    Dict,
    Point,
    LineString,
    Polygon,
    MultiPoint,
    MultiLineString,
    MultiPolygon,
]:
    """
    Creates a GeoJSON {@link Geometry} from a Geometry string type & coordinates.
    For GeometryCollection type use `helpers.geometryCollection`

    :param geom_type: one of "Point" | "LineString" | "Polygon" | "MultiPoint" | "MultiLineString" | "MultiPolygon"
    :param coordinates: array of coordinates [lng, lat]
    :param as_geojson: whether the return value should be a geojson
    :return: a GeoJSON geometry
    """

    if geom_type == "Point":
        geom = Point(coordinates)
    elif geom_type == "LineString":
        geom = LineString(coordinates)
    elif geom_type == "Polygon":
        geom = Polygon(coordinates)
    elif geom_type == "MultiPoint":
        geom = MultiPoint(coordinates)
    elif geom_type == "MultiLineString":
        geom = MultiLineString(coordinates)
    elif geom_type == "MultiPolygon":
        geom = MultiPolygon(coordinates)
    else:
        raise InvalidInput(error_code_messages["InvalidGeometry"](all_geometry_types))

    return geom.to_geojson() if as_geojson else geom


def point(
    coordinates: Sequence,
    properties: Dict = None,
    options: Dict = None,
    as_geojson: bool = True,
) -> Union[Point, Dict]:
    """
    Creates a Point Feature from a Position.

    :param coordinates: coordinates longitude, latitude position in degrees - Position
    :param properties: a dictionary of key-value pairs to add as properties
    :param options: an options dictionary:
        [options["bbox"] Bounding Box Array [west, south, east, north] associated with the Feature
        [options["id"] Identifier associated with the Feature
    :param as_geojson: whether the return value should be a geojson
    :return: a Point Feature
    """

    geom = Point(coordinates)

    return feature(geom, properties, options, as_geojson=as_geojson)


def points(
    coordinates: Sequence,
    properties: Dict = None,
    options: Dict = None,
    as_geojson: bool = True,
) -> Union[FeatureCollection, Dict]:
    """
    Creates a Point FeatureCollection from an Array of Point coordinates.

    :param coordinates: a list of Points - Position[]
    :param properties: a dictionary of key-value pairs to add as properties
    :param options: an options dictionary:
        [options["bbox"] Bounding Box Array [west, south, east, north] associated with the Feature
        [options["id"] Identifier associated with the Feature
    :param as_geojson: whether the return value should be a geojson
    :return: Point FeatureCollection
    """

    if not isinstance(coordinates, list):
        raise Exception("Coordinates must be a list")

    return feature_collection(
        list(map(lambda coord: point(coord, properties), coordinates)),
        options,
        as_geojson=as_geojson,
    )


def multi_point(
    coordinates: Sequence,
    properties: Dict = None,
    options: Dict = None,
    as_geojson: bool = True,
) -> Union[MultiPoint, Dict]:
    """
    Creates a MultiPoint Feature based on a coordinate array.
    Properties can be added optionally.

    :param coordinates: a list of Points - Position[]
    :param properties: a dictionary of key-value pairs to add as properties
    :param options: an options dictionary:
        [options["bbox"] Bounding Box Array [west, south, east, north] associated with the Feature
        [options["id"] Identifier associated with the Feature
    :param as_geojson: whether the return value should be a geojson
    :return: a MultiPoint feature
    """

    geom = MultiPoint(coordinates)

    return feature(geom, properties, options, as_geojson=as_geojson)


def line_string(
    coordinates: Sequence,
    properties: Dict = None,
    options: Dict = None,
    as_geojson: bool = True,
) -> Union[LineString, Dict]:
    """
    Creates a LineString Feature from an Array of Positions.

    :param coordinates: a list of Positions - Position[]
    :param properties: a dictionary of key-value pairs to add as properties
    :param options: an options dictionary:
        [options["bbox"] Bounding Box Array [west, south, east, north] associated with the Feature
        [options["id"] Identifier associated with the Feature
    :param as_geojson: whether the return value should be a geojson
    :return: a LineString feature
    """

    if not options:
        options = {}

    if not properties:
        properties = {}

    geom = LineString(coordinates)

    return feature(geom, properties, options, as_geojson=as_geojson)


def line_strings(
    coordinates: Sequence,
    properties: Dict = None,
    options: Dict = None,
    as_geojson: bool = True,
) -> Union[FeatureCollection, Dict]:
    """
    Creates a LineString FeatureCollection from an Array of LineString coordinates.

    :param coordinates: a list of a list of Positions - Position[][]
    :param properties: a dictionary of key-value pairs to add as properties
    :param options: an options dictionary:
        [options["bbox"] Bounding Box Array [west, south, east, north] associated with the Feature
        [options["id"] Identifier associated with the Feature
    :param as_geojson: whether the return value should be a geojson
    :return: LineString FeatureCollection
    """

    if not isinstance(coordinates, list):
        raise Exception("Coordinates_list must be a list")

    return feature_collection(
        list(map(lambda coord: line_string(coord, properties), coordinates)),
        options,
        as_geojson=as_geojson,
    )


def multi_line_string(
    coordinates: Sequence,
    properties: Dict = None,
    options: Dict = None,
    as_geojson: bool = True,
) -> Union[MultiLineString, Dict]:
    """
    Creates a MultiLineString Feature based on a coordinate array.
    Properties can be added optionally.

    :param coordinates: a list of a list of Positions - Position[][]
    :param properties: a dictionary of key-value pairs to add as properties
    :param options: an options dictionary:
        [options["bbox"] Bounding Box Array [west, south, east, north] associated with the Feature
        [options["id"] Identifier associated with the Feature
    :param as_geojson: whether the return value should be a geojson
    :return: a MultiLineString feature
    """

    geom = MultiLineString(coordinates)

    return feature(geom, properties, options, as_geojson=as_geojson)


def polygon(
    coordinates: Sequence,
    properties: Dict = None,
    options: Dict = None,
    as_geojson: bool = True,
) -> Union[Polygon, Dict]:
    """
    Creates a Polygon Feature from an Array of LinearRings.

    :param coordinates: a list of a list of Positions - Position[][]
    :param properties: a dictionary of key-value pairs to add as properties
    :param options: an options dictionary:
        [options["bbox"] Bounding Box Array [west, south, east, north] associated with the Feature
        [options["id"] Identifier associated with the Feature
    :param as_geojson: whether the return value should be a geojson
    :return: a Polygon Feature
    """

    if not options:
        options = {}

    if not properties:
        properties = {}

    geom = Polygon(coordinates)

    return feature(geom, properties, options, as_geojson=as_geojson)


def polygons(
    coordinates: Sequence,
    properties: Dict = None,
    options: Dict = None,
    as_geojson: bool = True,
) -> Union[FeatureCollection, Dict]:
    """
    Creates a Polygon FeatureCollection from an Array of Polygon coordinates.

    :param coordinates: an array of polygons - Position[][][]
    :param properties: a dictionary of key-value pairs to add as properties
    :param options: an options dictionary:
        [options["bbox"] Bounding Box Array [west, south, east, north] associated with the Feature
        [options["id"] Identifier associated with the Feature
    :param as_geojson: whether the return value should be a geojson
    :return: Polygon FeatureCollection
    """

    if not isinstance(coordinates, list):
        raise Exception("Coordinates_list must be a list")

    return feature_collection(
        list(map(lambda coords: polygon(coords, properties), coordinates)),
        options,
        as_geojson=as_geojson,
    )


def multi_polygon(
    coordinates: Sequence,
    properties: Dict = None,
    options: Dict = None,
    as_geojson: bool = True,
) -> Union[MultiPolygon, Dict]:
    """
    Creates a MultiPolygon Feature based on a coordinate array.
    Properties can be added optionally.

    :param coordinates: an array of polygons - Position[][][]
    :param properties: a dictionary of key-value pairs to add as properties
    :param options: an options dictionary:
        [options["bbox"] Bounding Box Array [west, south, east, north] associated with the Feature
        [options["id"] Identifier associated with the Feature
    :param as_geojson: whether the return value should be a geojson
    :return: a MultiPolygon feature
    """

    geom = MultiPolygon(coordinates)

    return feature(geom, properties, options, as_geojson=as_geojson)



from typing import Any, List, Sequence, Union

from turf.helpers import (
    Feature,
    LineString,
    MultiLineString,
    MultiPolygon,
    Point,
    Polygon,
)

from turf.helpers import line_string, point, polygon

from turf.boolean_point_in_polygon import boolean_point_in_polygon
from turf.boolean_point_on_line import boolean_point_on_line
from turf.invariant import get_coords_from_features, get_geometry_type
from turf.polygon_to_line import polygon_to_line
from turf.line_intersect import line_intersect

from turf.utils.error_codes import error_code_messages
from turf.utils.exceptions import InvalidInput


def boolean_disjoint(feature_1: Any, feature_2: Any) -> bool:
    """
    Returns true if the intersection of the two geometries is an empty set.

    :param feature_1: {GeoJSON} feature_1 any Feature or Geometry
    :param feature_2: {GeoJSON} feature_2 any Feature or Geometry
    :return: boolean True/False if features are disjoint
    """
    is_disjoint = True

    flat_feature_1 = flatten_feature(feature_1)
    flat_feature_2 = flatten_feature(feature_2)

    for flat_1 in flat_feature_1:
        for flat_2 in flat_feature_2:
            is_disjoint = disjoint(flat_1, flat_2)

            if not is_disjoint:
                return is_disjoint

    return is_disjoint


def disjoint(
    feature_1: List[Union[str, Sequence]], feature_2: List[Union[str, Sequence]]
) -> bool:
    """
    Returns true if the intersection of the two geometries is an empty set.

    :param feature_1: {List} a List with geometry type and coordinates
    :param feature_2: {List} a List with geometry type and coordinates
    :return: boolean True/False if features are disjoint
    """
    is_disjoint = True

    if feature_1[0] in ["Point"]:
        if feature_2[0] in ["Point"]:
            is_disjoint = feature_1[1] != feature_2[1]

        elif feature_2[0] in ["LineString"]:
            is_disjoint = not boolean_point_on_line(feature_1[1], feature_2[1])

        elif feature_2[0] in ["Polygon"]:
            is_disjoint = not boolean_point_in_polygon(feature_1[1], feature_2[1])

    elif feature_1[0] in ["LineString"]:
        if feature_2[0] in ["Point"]:
            is_disjoint = not boolean_point_on_line(feature_2[1], feature_1[1])

        elif feature_2[0] in ["LineString"]:
            is_disjoint = not is_line_on_line(feature_1[1], feature_2[1])

        elif feature_2[0] in ["Polygon"]:
            is_disjoint = not is_line_in_poly(feature_2[1], feature_1[1])

    elif feature_1[0] in ["Polygon"]:
        if feature_2[0] in ["Point"]:
            is_disjoint = not boolean_point_in_polygon(feature_2[1], feature_1[1])

        elif feature_2[0] in ["LineString"]:
            is_disjoint = not is_line_in_poly(feature_1[1], feature_2[1])

        elif feature_2[0] in ["Polygon"]:
            is_disjoint = not is_poly_in_poly(feature_2[1], feature_1[1])

    return is_disjoint


def is_line_on_line(feature_1: Sequence, feature_2: Sequence) -> bool:
    """
    Checks if two linestring feature intersects each other

    :param feature_1: Coordinates of linestring feature 1
    :param feature_2: Coordinates of linestring feature 2
    :return: bool if there is an intersection
    """
    if isinstance(feature_1, list):
        feature_1 = line_string(feature_1)

    if isinstance(feature_2, list):
        feature_2 = line_string(feature_2)

    intersects = line_intersect(feature_1, feature_2)

    if len(intersects["features"]) >= 1:
        return True

    return False


def is_line_in_poly(feature_1: Sequence, feature_2: Sequence) -> bool:
    """
    Checks if a linestring feature is inside or intersects a polygon feature

    :param feature_1: Coordinates of polygon feature
    :param feature_2: Coordinates of linestring feature
    :return: bool if there is an intersection
    """
    feature_1_line = polygon_to_line(polygon(feature_1))

    if is_line_on_line(feature_2, feature_1_line):
        return True

    for coord in feature_2:
        if boolean_point_in_polygon(coord, feature_1):
            return True

    return False


def is_poly_in_poly(feature_1: Sequence, feature_2: Sequence) -> bool:
    """
    Checks if polygon feature_1 is inside polygon feature_2 and either way
    See http://stackoverflow.com/a/4833823/1979085

    :param feature: Coordinates of polygon feature 1
    :param feature: Coordinates of polygon feature 1
    :return: bool if there is an intersection
    """
    feature_1_line = polygon_to_line(polygon(feature_1))
    feature_2_line = polygon_to_line(polygon(feature_2))

    for coord1 in feature_1_line["geometry"]["coordinates"]:
        if boolean_point_in_polygon(coord1, feature_2):
            return True

    for coord2 in feature_2_line["geometry"]["coordinates"]:
        if boolean_point_in_polygon(coord2, feature_1):
            return True

    if is_line_on_line(feature_1_line, feature_2_line):
        return True

    return False


def flatten_feature(feature: Any) -> List[Union[str, Sequence]]:
    """
    Takes any feature and return the simple geometry type with coordinates.
    A MultiPoint will be flatten to Point, MultiLineString to LineString and
    MultiPolygon to Polygon

    :param feature: {GeoJSON} feature any Feature or Geometry
    :return: List of geometry type and coordinate sequence
    """
    feature_coords = get_coords_from_features(feature)
    feature_geometry = get_geometry_type(feature)

    if isinstance(feature_geometry, (list, tuple)):
        feature_geometry = feature_geometry[0]

    if feature_geometry in ["MultiPoint", "MultiLineString", "MultiPolygon"]:
        if feature_geometry == "MultiPoint":
            feature_geometry = "Point"

        elif feature_geometry == "MultiLineString":
            feature_geometry = "LineString"

        elif feature_geometry == "MultiPolygon":
            feature_geometry = "Polygon"

        flat_feature = [[feature_geometry, coords] for coords in feature_coords]

    else:
        flat_feature = [[feature_geometry, feature_coords]]

    return flat_feature



from typing import Dict, List, Union

from turf.distance import distance
from turf.bbox import bbox
from turf.boolean_intersects import boolean_intersects
from turf.helpers import feature_collection, polygon, FeatureCollection


def rectangle_grid(
    bbox: List[float],
    cell_width: Union[int, float],
    cell_height: Union[int, float],
    options: Dict = {},
) -> FeatureCollection:
    """
    Creates a grid of rectangles from a bounding box, Feature or FeatureCollection.

    :param bbox: Array extent in [minX, minY, maxX, maxY] order
    :param cell_width: of each cell, in units
    :param cell_height: of each cell, in units

    :param options: Optional parameters
        [options["units"]]: units ("degrees", "radians", "miles", "kilometers")
                            of the given cell_width and cell_height
        [options["mask"]]: if passed a Polygon or MultiPolygon here,
                           the grid Points will be created only inside it
        [options["properties"]]: passed to each point of the grid

    :returns: FeatureCollection of a grid of polygons
    """
    if not isinstance(options, dict):
        options = {}

    results = []
    west = bbox[0]
    south = bbox[1]
    east = bbox[2]
    north = bbox[3]

    x_fraction = cell_width / (distance([west, south], [east, south], options))
    cell_width_deg = x_fraction * (east - west)
    y_fraction = cell_height / (distance([west, south], [west, north], options))
    cell_height_deg = y_fraction * (north - south)

    # rows & columns
    bbox_width = east - west
    bbox_height = north - south
    columns = int(bbox_width // cell_width_deg)
    rows = int(bbox_height // cell_height_deg)

    # if the grid does not fill the bbox perfectly, center it.
    delta_x = (bbox_width - columns * cell_width_deg) / 2
    delta_y = (bbox_height - rows * cell_height_deg) / 2

    # iterate over columns & rows
    current_x = west + delta_x
    for _ in range(columns):
        current_y = south + delta_y
        for _ in range(rows):
            cell_poly = polygon(
                [
                    [
                        [current_x, current_y],
                        [current_x, current_y + cell_height_deg],
                        [current_x + cell_width_deg, current_y + cell_height_deg],
                        [current_x + cell_width_deg, current_y],
                        [current_x, current_y],
                    ]
                ],
                options.get("properties", {}),
            )

            if "mask" in options:
                if boolean_intersects(options["mask"], cell_poly):
                    results.append(cell_poly)
            else:
                results.append(cell_poly)

            current_y += cell_height_deg

        current_x += cell_width_deg

    return feature_collection(results)



from typing import Any, List, Sequence, Union

from turf.helpers import point, polygon

from turf.bbox import bbox

# from turf.boolean_disjoint._boolean_disjoint import flatten_feature
from turf.boolean_point_in_polygon import boolean_point_in_polygon
from turf.boolean_point_on_line import boolean_point_on_line
from turf.invariant import get_coords_from_features, get_geometry_type
from turf.line_intersect import line_intersect
from turf.midpoint import midpoint
from turf.polygon_to_line import polygon_to_line


def boolean_within(feature_1: Any, feature_2: Any) -> bool:
    """
    Returns true if the first geometry is completely within the second geometry.

    The interiors of both geometries must intersect and, the interior and
    boundary of the primary (geometry a) must not intersect
    the exterior of the secondary (geometry b).

    :param feature_1: {GeoJSON} feature_1 any Feature or Geometry
    :param feature_2: {GeoJSON} feature_2 any Feature or Geometry

    :return: boolean True/False if feature 1 is within feature 2
    """
    feature_1 = get_features(feature_1)
    feature_2 = get_features(feature_2)

    return check_within(feature_1, feature_2)


def check_within(
    feature_1: List[Union[str, Sequence]], feature_2: List[Union[str, Sequence]]
) -> bool:
    """
    Returns true if the first geometry is completely within the second geometry

    :param feature_1: {List} a List with geometry type and coordinates
    :param feature_2: {List} a List with geometry type and coordinates

    :return: boolean True/False if feature 1 is within feature 2
    """
    is_within = False

    if feature_1[0] in ["Point"]:
        if feature_2[0] in ["Point", "MultiPoint"]:
            is_within = boolean_point_on_point(feature_1[1], feature_2[1])

        elif feature_2[0] in ["LineString", "MultiLineString"]:
            is_within = boolean_point_on_line(
                feature_1[1], feature_2[1], {"ignoreEndVertices": True}
            )

        elif feature_2[0] in ["Polygon", "MultiPolygon"]:
            is_within = boolean_point_in_polygon(
                feature_1[1], feature_2[1], {"ignoreBoundary": True}
            )

    if feature_1[0] in ["MultiPoint"]:
        if feature_2[0] in ["MultiPoint"]:
            is_within = all(
                boolean_point_on_point(coords_1, feature_2[1])
                for coords_1 in feature_1[1]
            )

        elif feature_2[0] in ["LineString", "MultiLineString"]:
            is_within = is_multipoint_on_linestring(feature_1, feature_2)

        elif feature_2[0] in ["Polygon", "MultiPolygon"]:
            is_within = is_multipoint_on_polygon(feature_1, feature_2)

    elif feature_1[0] in ["LineString"]:
        if feature_2[0] in ["LineString"]:
            is_within = is_line_on_line(feature_1[1], feature_2[1])

        if feature_2[0] in ["MultiLineString"]:
            is_within = is_line_on_multiline(feature_1[1], feature_2[1])

        elif feature_2[0] in ["Polygon"]:
            is_within = is_line_in_poly(feature_1[1], feature_2[1])

        elif feature_2[0] in ["MultiPolygon"]:
            is_within = is_line_in_multipoly(feature_1[1], feature_2[1])

    elif feature_1[0] in ["MultiLineString"]:
        if feature_2[0] in ["MultiLineString"]:
            is_within = all(
                is_line_on_multiline(coords_1, feature_2[1])
                for coords_1 in feature_1[1]
            )

        elif feature_2[0] in ["Polygon", "MultiPolygon"]:
            is_within = all(
                is_line_in_poly(coords_1, feature_2[1]) for coords_1 in feature_1[1]
            )

    elif feature_1[0] in ["Polygon"]:
        if feature_2[0] in ["Polygon"]:
            is_within = is_poly_in_poly(feature_1[1], feature_2[1])

        if feature_2[0] in ["MultiPolygon"]:
            is_within = is_poly_in_multipoly(feature_1[1], feature_2[1])

    elif feature_1[0] in ["MultiPolygon"]:
        if feature_2[0] in ["MultiPolygon"]:
            is_within = all(
                is_poly_in_multipoly(coords_1, feature_2[1])
                for coords_1 in feature_1[1]
            )

    return is_within


def boolean_point_on_point(feature_1: Sequence, feature_2: Sequence) -> bool:
    """
    Checks if feature_1 point is equal feature_2 point

    :param feature_1: Coordinates of point feature 1
    :param feature_2: Coordinates of point feature 2

    :return: boolean True/False if feature 1 is equal feature 2
    """

    return any(feature_1 == coords_2 for coords_2 in feature_2)


def is_multipoint_on_linestring(feature_1: Sequence, feature_2: Sequence) -> bool:
    """
    Checks if feature_1 multipoint feature is in feature_2 polygon
    returns False, if all multipoints are on the boundary

    :param feature_1: Coordinates of multipoint feature
    :param feature_2: Coordinates of polygon feature 2

    :return: boolean True/False if feature 1 is within feature 2
    """

    points_on_line = False

    points_on_line = all(
        boolean_point_on_line(coords_1, feature_2[1]) for coords_1 in feature_1[1]
    )

    if not points_on_line:
        return points_on_line

    points_on_line = any(
        boolean_point_on_line(coords_1, feature_2[1], {"ignoreEndVertices": True})
        for coords_1 in feature_1[1]
    )

    return points_on_line


def is_multipoint_on_polygon(feature_1: Sequence, feature_2: Sequence) -> bool:
    """
    Checks if feature_1 multipoint feature is in feature_2 polygon
    returns False, if all multipoints are on the boundary

    :param feature_1: Coordinates of multipoint feature
    :param feature_2: Coordinates of polygon feature 2

    :return: boolean True/False if feature 1 is within feature 2
    """
    points_on_poly = False

    points_on_poly = all(
        boolean_point_in_polygon(coords_1, feature_2[1]) for coords_1 in feature_1[1]
    )

    if not points_on_poly:
        return points_on_poly

    points_on_poly = any(
        boolean_point_in_polygon(coords_1, feature_2[1], {"ignoreBoundary": True})
        for coords_1 in feature_1[1]
    )

    return points_on_poly


def is_line_on_line(feature_1: Sequence, feature_2: Sequence) -> bool:
    """
    Checks if feature_1 linestring feature is in feature_2 linestring

    :param feature_1: Coordinates of linestring feature 1
    :param feature_2: Coordinates of linestring feature 2

    :return: boolean True/False if feature 1 is within feature 2
    """

    line_on_line = False

    for coords in feature_1:
        line_on_line = boolean_point_on_line(coords, feature_2)
        if not line_on_line:
            break

    return line_on_line


def is_line_on_multiline(feature_1: Sequence, feature_2: Sequence) -> bool:
    """
    Checks if feature_1 linestring feature is in any linestring of feature_2
    multilinestring

    :param feature_1: Coordinates of linestring feature 1
    :param feature_2: Coordinates of multilinestring feature 2

    :return: boolean True/False if feature 1 is within feature 2
    """
    return any(is_line_on_line(feature_1, coords_2) for coords_2 in feature_2)


def is_line_in_poly(feature_1: Sequence, feature_2: Sequence) -> bool:
    """
    Checks if feature_1 linestring feature is in feature_2 polygon

    :param feature_1: Coordinates of linestring feature 1
    :param feature_2: Coordinates of polygon feature 2

    :return: boolean True/False if feature 1 is within feature 2
    """
    line_in_poly = False

    line_bbox = bbox(feature_1)
    poly_bbox = bbox(feature_2)

    if not bbox_overlap(poly_bbox, line_bbox):
        return False

    for i in range(len(feature_1) - 1):
        if not boolean_point_in_polygon(feature_1[i], feature_2):
            return False

        if not line_in_poly:
            line_in_poly = boolean_point_in_polygon(
                feature_1[i], feature_2, {"ignoreBoundary": True}
            )

        if not line_in_poly:
            mid = midpoint(point(feature_1[i]), point(feature_1[i + 1]))
            line_in_poly = boolean_point_in_polygon(
                mid, feature_2, {"ignoreBoundary": True}
            )

    return line_in_poly


def is_line_in_multipoly(feature_1: Sequence, feature_2: Sequence) -> bool:
    """
    Checks if feature_1 line feature is in any feature_2 multipolygon

    :param feature_1: Coordinates of line feature 1
    :param feature_2: Coordinates of multipolygon feature 2

    :return: boolean True/False if any feature 1 is within any feature 2
    """

    return any(is_line_in_poly(feature_1, coords_2) for coords_2 in feature_2)


def is_poly_in_poly(feature_1: Sequence, feature_2: Sequence) -> bool:
    """
    Checks if feature_1 polygon feature is in feature_2 polygon

    :param feature_1: Coordinates of polygon feature 1
    :param feature_2: Coordinates of polygon feature 2

    :return: boolean True/False if feature 1 is within feature 2
    """
    poly_bbox_1 = bbox(feature_1)
    poly_bbox_2 = bbox(feature_2)

    if not bbox_overlap(poly_bbox_2, poly_bbox_1):
        return False

    feature_1 = polygon_to_line(polygon(feature_1))
    line_coords = get_coords_from_features(feature_1)

    for coords in line_coords:
        if not boolean_point_in_polygon(coords, feature_2):
            return False

    return True


def is_poly_in_multipoly(feature_1: Sequence, feature_2: Sequence) -> bool:
    """
    Checks if feature_1 polygon feature is in feature_2 multipolygon

    :param feature_1: Coordinates of polygon feature 1
    :param feature_2: Coordinates of multipolygon feature 2

    :return: boolean True/False if any feature 1 is within feature 2
    """

    return any(is_poly_in_poly(feature_1, coords_2) for coords_2 in feature_2)


def get_features(feature: Any) -> List[Union[str, Sequence]]:
    """
    Takes any feature and returns the geometry type with coordinates.

    :param feature: {GeoJSON} feature any Feature or Geometry

    :return: List of geometry type and coordinate sequence
    """
    feature_coords = get_coords_from_features(feature)
    feature_geometry = get_geometry_type(feature)

    if isinstance(feature_geometry, (list, tuple)):
        feature_geometry = feature_geometry[0]

    return [feature_geometry, feature_coords]


def bbox_overlap(bbox_1: Sequence, bbox_2: Sequence) -> bool:
    """
    Takes two bounding boxes and tests, if bbox_1 is inside bbox2

    :param bbox_1: {List} bounding box of feature 1
    :param bbox_2: {List} bounding box of feature 2

    :return: boolean True/False if bbox 1 is within bbox 2
    """
    if (bbox_1[0] > bbox_2[0]) or (bbox_1[1] > bbox_2[1]):
        return False
    if (bbox_1[2] < bbox_2[2]) or (bbox_1[3] < bbox_2[3]):
        return False

    return True



from typing import Any

from turf.boolean_disjoint import boolean_disjoint


def boolean_intersects(feature_1: Any, feature_2: Any) -> bool:
    """
    Returns true if the two geometries intersect.

    :param feature_1: {GeoJSON} feature_1 any Feature or Geometry
    :param feature_2: {GeoJSON} feature_2 any Feature or Geometry
    :return: boolean True/False if features intersect
    """
    return not boolean_disjoint(feature_1, feature_2)



from typing import Dict, List, Sequence, TypeVar, Union

from turf.bbox import bbox
from turf.explode import explode
from turf.nearest_point import nearest_point

from turf.helpers import (
    all_geometry_types,
    FeatureCollection,
    MultiPolygon,
    Point,
    Polygon,
)
from turf.helpers import feature, feature_collection, geometry, point, polygon
from turf.helpers import Feature, FeatureCollection, Geometry
from turf.invariant import get_coords_from_features, get_geometry_type
from turf.utils.error_codes import error_code_messages
from turf.utils.exceptions import InvalidInput

GeoJson = TypeVar("GeoJson", Dict, Feature, FeatureCollection, Geometry)
PointFeature = TypeVar("PointFeature", Dict, Point, Sequence)
PolygonFeature = TypeVar("PolygonFeature", Dict, Polygon, MultiPolygon)


def polygon_tangents(
    start_point: PointFeature, polygon: PolygonFeature
) -> FeatureCollection:
    """Finds the tangents of a {Polygon or(MultiPolygon} from a {Point}.

    more:
    http://geomalgorithms.com/a15-_tangents.html

    :param point: point [lng, lat] or Point feature to calculate the tangent points from
    :param polygon: polygon to get tangents from

    :return:
        Feature Collection containing the two tangent points
    """
    point_features = []

    point_coord = get_coords_from_features(start_point, ("Point",))
    polygon_coords = get_coords_from_features(polygon, ("Polygon", "MultiPolygon"))

    geometry_type = get_geometry_type(polygon)

    if isinstance(geometry_type, str):
        geometry_type = [geometry_type]
        polygon_coords = [polygon_coords]

    box = bbox(polygon)

    near_point_index = 0
    near_point = False

    # If the point lies inside the polygon bbox then it's a bit more complicated
    # points lying inside a polygon can reflex angles on concave polygons
    if (
        (point_coord[0] > box[0])
        and (point_coord[0] < box[2])
        and (point_coord[1] > box[1])
        and (point_coord[1] < box[3])
    ):
        near_point = nearest_point(start_point, explode(polygon))
        near_point_index = near_point["properties"]["featureIndex"]

    for geo_type, poly_coords in zip(geometry_type, polygon_coords):
        if geo_type == "Polygon":
            tangents = process_polygon(
                poly_coords, point_coord, near_point, near_point_index
            )

        # bruteforce approach
        # calculate both tangents for each polygon
        # define all tangents as a new polygon and calculate tangetns out of those coordinates
        elif geo_type == "MultiPolygon":
            multi_tangents = []

            for poly_coord in poly_coords:
                tangents = process_polygon(
                    poly_coord, point_coord, near_point, near_point_index
                )
                multi_tangents.extend(tangents)

            tangents = process_polygon(
                [multi_tangents], point_coord, near_point, near_point_index
            )

        r_tangents = tangents[0]
        l_tangents = tangents[1]

        point_features.extend([point(r_tangents), point(l_tangents)])

    return feature_collection(point_features)


def process_polygon(
    polygon_coords: Sequence,
    point_coord: Sequence,
    near_point: Union[Point, None],
    near_point_index: int,
) -> Sequence:
    """Prepares a polygon to calculate the tangents

    :param polygon_coords: point [lng, lat] or Point feature to calculate the tangent points from
    :param point_coord: point [lng, lat]
    :param near_point: nearest point [lng, lat] on polygon towards view point
                         in case the view point lies inside the polygon
    :param near_point_index: index of neareast point on polygon
                        in case the view point lies inside the polygon

    :return:
        List of tangents coordinates [upper and lower]
    """
    r_tangents = polygon_coords[0][near_point_index]
    l_tangents = polygon_coords[0][0]

    if near_point:
        if near_point["geometry"]["coordinates"][1] < point_coord[1]:
            l_tangents = polygon_coords[0][near_point_index]

    tangents = calculate_tangents(
        polygon_coords[0],
        point_coord,
        r_tangents,
        l_tangents,
    )

    return tangents


def calculate_tangents(
    poly_coords: Sequence,
    point_coord: Sequence,
    r_tangents: Sequence,
    l_tangents: Sequence,
) -> Sequence:
    """Calculate the upper and lower tangents from the polygon by
    comparison with each neighbor coordinate in the polygon

    :param polygon_coords: point [lng, lat] or Point feature to calculate the tangent points from
    :param point_coord: point [lng, lat]
    :param r_tangents: current rightmost tangents point [lng, lat] of the polygon
    :param l_tangents: current leftmost tangents point [lng, lat] of the polygon

    :return:
        List of tangents coordinates [upper and lower]
    """
    edge_next = None
    edge_prev = is_left(
        poly_coords[0],
        poly_coords[len(poly_coords) - 1],
        point_coord,
    )

    for i in range(1, len(poly_coords) + 1):
        current_poly_coord = poly_coords[i - 1]

        if i == len(poly_coords):
            next_poly_coord = poly_coords[0]
        else:
            next_poly_coord = poly_coords[i]

        edge_next = is_left(current_poly_coord, next_poly_coord, point_coord)

        if (edge_prev <= 0) and (edge_next > 0):
            if not is_below(point_coord, current_poly_coord, r_tangents):
                r_tangents = current_poly_coord

        elif (edge_prev > 0) and (edge_next <= 0):
            if not is_above(point_coord, current_poly_coord, l_tangents):
                l_tangents = current_poly_coord

        edge_prev = edge_next

    return [r_tangents, l_tangents]


def is_above(point1: Sequence, point2: Sequence, point3: Sequence) -> bool:
    """Checks if point 1 is above point2 and point 3

    :param geojson: geojson or Feature
    :param geojson_types: string of the supposed feature tpye

    :return:
        Feature or FeatureCollection of the incoming geojson
    """
    return is_left(point1, point2, point3) > 0


def is_below(point1: Sequence, point2: Sequence, point3: Sequence) -> bool:
    """Checks if point 1 is below point2 and point 3

    :param geojson: geojson or Feature
    :param geojson_types: string of the supposed feature tpye

    :return:
        Feature or FeatureCollection of the incoming geojson
    """
    return is_left(point1, point2, point3) < 0


def is_left(point1: Sequence, point2: Sequence, point3: Sequence) -> float:
    """Calculates if point 1 is left of point2 and point 3

    :param geojson: geojson or Feature
    :param geojson_types: string of the supposed feature tpye

    :return:
        Feature or FeatureCollection of the incoming geojson
    """
    return (point2[0] - point1[0]) * (point3[1] - point1[1]) - (
        point3[0] - point1[0]
    ) * (point2[1] - point1[1])



from typing import Dict, List, Union
import math

from turf.distance import distance
from turf.bbox import bbox
from turf.boolean_intersects import boolean_intersects
from turf.helpers import feature_collection, polygon, FeatureCollection


def hex_grid(
    bbox: List[float],
    cell_side: Union[int, float],
    options: Dict = {},
) -> FeatureCollection:
    """
    Takes a bounding box and the diameter of the cell and returns a FeatureCollection of flat-topped
    hexagons or triangles aligned in an "odd-q" vertical grid as
    described in [Hexagonal Grids](http://www.redblobgames.com/grids/hexagons/).

    :param bbox: Array extent in [minX, minY, maxX, maxY] order
    :param n_cells: length of the side of the the hexagons or triangles, in units. It will also coincide with the
                    radius of the circumcircle of the hexagons
    :param options: Optional parameters
        [options["units"]]: units ("degrees", "radians", "miles", "kilometers")
                            of the given cell_width and cell_height
        [options["mask"]]: if passed a Polygon or MultiPolygon here,
                           the grid Points will be created only inside it
        [options["properties"]]: passed to each point of the grid
        [options["triangles"]]: whether to return as triangles instead of hexagons

    :returns: FeatureCollection of a grid of polygons
    """
    if not isinstance(options, dict):
        options = {}

    has_triangles = options.get("triangles", None)

    results = []
    west = bbox[0]
    south = bbox[1]
    east = bbox[2]
    north = bbox[3]

    center_y = (south + north) / 2
    center_x = (west + east) / 2

    x_fraction = (cell_side * 2) / (
        distance([west, center_y], [east, center_y], options)
    )
    cell_width_deg = x_fraction * (east - west)
    y_fraction = (
        cell_side * 2 / (distance([center_x, south], [center_x, north], options))
    )
    cell_height_deg = y_fraction * (north - south)
    radius = cell_width_deg / 2

    hex_width = radius * 2
    hex_height = math.sqrt(3) / 2 * cell_height_deg

    # rows & columns
    bbox_width = east - west
    bbox_height = north - south

    x_interval = 3 / 4 * hex_width
    y_interval = hex_height

    x_span = (bbox_width - hex_width) / (hex_width - radius / 2)
    x_count = int(x_span)

    x_adjust = (
        ((x_count * x_interval - radius / 2) - bbox_width) / 2
        - radius / 2
        + x_interval / 2
    )

    y_count = int((bbox_height - hex_height) / hex_height)
    y_adjust = (bbox_height - y_count * hex_height) / 2

    has_offset_y = (y_count * hex_height - bbox_height) > (hex_height / 2)

    if has_offset_y:
        y_adjust -= hex_height / 4

    cosines = []
    sines = []

    for i in range(6):
        angle = 2 * math.pi / 6 * i
        cosines.append(math.cos(angle))
        sines.append(math.sin(angle))

    results = []

    for x in range(x_count + 1):
        for y in range(y_count + 1):
            is_odd = x % 2 == 1

            if (y == 0) and is_odd:
                continue

            if (y == 0) and has_offset_y:
                continue

            center_x = x * x_interval + west - x_adjust
            center_y = y * y_interval + south + y_adjust

            if is_odd:
                center_y -= hex_height / 2

            if has_triangles:
                triangles = hex_triangles(
                    [center_x, center_y],
                    cell_width_deg / 2,
                    cell_height_deg / 2,
                    options.get("properties", {}).copy(),
                    cosines,
                    sines,
                )

                for triangle in triangles:
                    if "mask" in options:
                        if boolean_intersects(options["mask"], triangle):
                            results.append(triangle)
                    else:
                        results.append(triangle)

            else:
                hex = hexagon(
                    [center_x, center_y],
                    cell_width_deg / 2,
                    cell_height_deg / 2,
                    options.get("properties", {}).copy(),
                    cosines,
                    sines,
                )

                if "mask" in options:
                    if boolean_intersects(options["mask"], hex):
                        results.append(hex)
                else:
                    results.append(hex)

    return feature_collection(results)


def hexagon(center, rx, ry, properties, cosines, sines):
    vertices = []
    for i in range(6):
        x = center[0] + rx * cosines[i]
        y = center[1] + ry * sines[i]
        vertices.append([x, y])

    vertices.append(vertices[0])
    return polygon([vertices], properties)


def hex_triangles(center, rx, ry, properties, cosines, sines):
    triangles = []
    for i in range(6):
        vertices = []
        vertices.append(center)
        vertices.append([center[0] + rx * cosines[i], center[1] + ry * sines[i]])
        vertices.append(
            [center[0] + rx * cosines[(i + 1) % 6], center[1] + ry * sines[(i + 1) % 6]]
        )
        vertices.append(center)
        triangles.append(polygon([vertices], properties))

    return triangles



from typing import Any, Sequence, Callable, Set, List, Tuple, Union

from turf.helpers import (
    Feature,
    Point,
    LineString,
    MultiPoint,
    MultiLineString,
    Polygon,
    MultiPolygon,
    FeatureCollection,
    get_input_dimensions,
)
from turf.utils.exceptions import InvalidInput
from turf.utils.error_codes import error_code_messages
from turf.utils.helpers import dimensions

allowed_types_points = ["Point", "MultiPoint"]
allowed_types_default = [
    "Point",
    "MultiPoint",
    "LineString",
    "MultiLineString",
    "Polygon",
    "MultiPolygon",
]
allowed_features_default = (
    Point,
    MultiPoint,
    LineString,
    MultiLineString,
    Polygon,
    MultiPolygon,
)


def _process_list_input(
    allowed_types: Sequence,
    input_value: Sequence,
    callback: Callable,
    raise_exception: bool = True,
) -> List:
    """
    Retrieves the coords from lists, or raises an exception if invalid.

    :param allowed_types: allowed Feature types
    :param input_value: list input
    :param callback: method to call recursively
    :param raise_exception: if an exception should be raised or if it should be silent
    :return: list with extracted coords
    """
    dim = get_input_dimensions(input_value)

    allowed_dimensions = {dimensions.get(type_, None) for type_ in allowed_types}

    allowed_class_types = (
        *[eval(allowed_type) for allowed_type in allowed_types],
        dict,
    )

    if dim == 1 and all(isinstance(el, allowed_class_types) for el in input_value):
        return list(map(lambda geom: callback(geom, allowed_types), input_value))

    return _check_input_in_allowed_dimensions(
        input_value, dim, allowed_types, allowed_dimensions, raise_exception
    )


def _check_input_in_allowed_dimensions(
    input_value: Sequence,
    dim: int,
    allowed_types: Sequence,
    allowed_dimensions: Set,
    raise_exception: bool = True,
) -> Union[List, Sequence]:
    """
    :param input_value: list input
    :param dim: number of dimensions of input
    :param allowed_types: allowed Feature types
    :param allowed_dimensions: allowed dimensions for specified allowed types
    :param raise_exception: if an exception should be raised or if it should be silent
    :return: list with extracted coords
    """
    if dim in allowed_dimensions:
        if dim == 1 and len(input_value) < 2:
            raise InvalidInput(error_code_messages["InvalidPointInput"])

        return input_value
    else:
        if raise_exception:
            raise InvalidInput(
                error_code_messages["InvalidGeometry"](
                    [type_ for type_ in allowed_types if type_ in allowed_types_default]
                )
            )
        return []


def get_coords_from_geometry(
    geometry: Any, allowed_types: Sequence = None, raise_exception: bool = True
) -> List:
    """
    Retrieves coords from a given Geometry. Geometry must be a GeoJSON,
    a Geometry object or a list of coordinates, otherwise it raises an exception.

    :param geometry: Any input value(s)
    :param allowed_types: allowed Feature types
    :param raise_exception: if an exception should be raised or if it should be silent
    :return: list with extracted coords
    """

    if not allowed_types:
        allowed_types = allowed_types_default

    if isinstance(geometry, (list, tuple)):
        return _process_list_input(
            allowed_types, geometry, get_coords_from_geometry, raise_exception
        )

    if isinstance(geometry, (Feature, dict)):
        if geometry.get("type") == "Feature":
            return get_coords_from_geometry(geometry.get("geometry", {}), allowed_types)

    allowed_class_types = [
        *[eval(allowed_type) for allowed_type in allowed_types],
        dict,
    ]

    if any(
        isinstance(geometry, allowed_type_class)
        and geometry.get("type", "") in allowed_types
        for allowed_type_class in allowed_class_types
    ):
        return geometry.get("coordinates", [])

    if raise_exception:
        raise InvalidInput(error_code_messages["InvalidGeometry"](allowed_types))
    else:
        return []


def get_geometry_from_features(features: Any, allowed_types: Sequence = None) -> List:
    """
    Retrieves Geometries from Features. Features must be a GeoJSON,
    a Feature object or a list of coordinates, otherwise it raises an exception.

    :param features: Any input value(s)
    :param allowed_types: allowed Feature types
    :return: list with extracted coords
    """

    if not allowed_types:
        allowed_types = allowed_types_default

    if isinstance(features, (list, tuple)):
        return _process_list_input(
            [*allowed_types, "Feature"], features, get_geometry_from_features
        )

    if isinstance(features, (FeatureCollection, dict)):
        if features.get("type") == "FeatureCollection":
            return list(
                map(
                    lambda feature: feature.get("geometry", {}),
                    features.get("features", []),
                )
            )

    if isinstance(features, (Feature, dict)):
        if features.get("type", "") == "Feature":
            if features.get("geometry", {}).get("type", "") in allowed_types:
                return features.get("geometry", {})

    if isinstance(
        features, (*[eval(allowed_type) for allowed_type in allowed_types], dict)
    ):
        if features.get("type", "") in allowed_types:
            return features

    raise InvalidInput(error_code_messages["InvalidGeometry"](allowed_types))


def get_geometry_type(
    features: Any, allowed_types: Sequence = None
) -> Union[str, Tuple[str]]:
    """
    Gets the Geometry type from Features. Features must be a GeoJSON,
    a Feature object, FeatureCollection a Dictionary, List or a Tuple, otherwise it raises an exception.

    :param features: Any input value(s)
    :param allowed_types: allowed Feature types
    :return: str if one extracted geometry_type else tuple
    """
    if not allowed_types:
        allowed_features = [allowed_features_default, allowed_types_default]

    else:
        allowed_features = [
            tuple(
                [
                    feature
                    for feature in allowed_features_default
                    if feature.__name__ in allowed_types
                ]
            ),
            allowed_types,
        ]

    features = get_geometry_from_features(features, allowed_features[1])

    if isinstance(features, (dict, *allowed_features[0])):
        geometry_type = _get_geometry_type_from_feature(features, allowed_features)

    elif isinstance(features, (list, tuple)):
        geometry_type = _get_geometry_type_from_list(features, allowed_features)

    else:
        raise InvalidInput(error_code_messages["InvalidGeometry"](allowed_features[1]))

    if len(geometry_type) == 1:
        geometry_type = geometry_type[0]

    return geometry_type


def _get_geometry_type_from_feature(
    features: Any, allowed_features: List[Union[Tuple, Sequence]]
) -> str:
    """
    Gets the Geometry type from a Features. Features must be a GeoJSON,
    a Feature object, FeatureCollection or a Dictionary, otherwise it raises an exception.

    :param features: Any input value(s) except a list
    :return: tuple with extracted geometry types
    """
    geometry_type = features.get("type", None)

    if geometry_type not in allowed_features[1]:
        raise InvalidInput(error_code_messages["InvalidGeometry"](allowed_features[1]))

    return geometry_type


def _get_geometry_type_from_list(
    features: List, allowed_features: List[Union[Tuple, Sequence]]
) -> Tuple[str]:
    """
    Gets the Geometry type from a List, otherwise it raises an exception.

    :param features: input feature as a list
    :return: tuple with extracted geometry types
    """
    geometry_type = tuple()

    n_dim = get_input_dimensions(features)

    if n_dim == 1 and all(
        isinstance(el, (dict, *allowed_features[0])) for el in features
    ):
        return tuple(
            map(
                lambda geom: _get_geometry_type_from_feature(geom, allowed_features),
                features,
            )
        )

    elif all(isinstance(el, (list, tuple, int, float)) for el in features):
        feature_type = [
            k for k, v in dimensions.items() if v == n_dim and k in allowed_features[1]
        ]
        if len(feature_type) == 1:
            geometry_type += (feature_type[0],)

        else:
            raise InvalidInput(
                error_code_messages["InvalidGeometry"](allowed_features[1])
            )

    else:
        raise InvalidInput(error_code_messages["InvalidGeometry"](allowed_features[1]))

    return geometry_type


def get_coords_from_features(features: Any, allowed_types: Sequence = None) -> List:
    """
    Retrieves coords from Features. Features must be a GeoJSON,
    a Feature object or a list of coordinates, otherwise it raises an exception.

    :param features: Any input value(s)
    :param allowed_types: allowed Feature types
    :return: list with extracted coords
    """

    if not allowed_types:
        allowed_types = allowed_types_default

    if isinstance(features, (FeatureCollection, dict)):
        if features.get("type") == "FeatureCollection":
            return list(
                map(
                    lambda feature: get_coords_from_geometry(
                        feature.get("geometry", {}), allowed_types
                    ),
                    features.get("features", []),
                )
            )

    return get_coords_from_geometry(features, allowed_types)



from math import cos, fmod, log, pi, sin, tan
from typing import Dict, List, Sequence, Union

from turf.helpers import convert_length, degrees_to_radians
from turf.helpers import earth_radius
from turf.helpers import point, Point
from turf.invariant import get_coords_from_features


def rhumb_destination(features: Dict, options: Dict = None) -> Point:
    """
    Returns the destination {Point} having travelled the given distance along a
    Rhumb line from the origin Point with the (varant) given bearing.

    # https://en.wikipedia.org/wiki/Rhumb_line

    :param features: any GeoJSON feature or feature collection
    :param properties: specification to calculate the rhumb line
        [options["distance"]=100] distance from the starting point
        [options["bearing"]=180] varant bearing angle ranging from -180 to 180 degrees from north
        [options["units"]=kilometers] units: specifies distance (can be degrees, radians, miles, or kilometers)

    :param options: optional parameters also be part of features["properties"]
        [options["units"]={}] can be degrees, radians, miles, or kilometers
        [options["properties"]={}] Translate GeoJSON Properties to Point
        [options["id"]={}] Translate GeoJSON Id to Point

    :return: a FeatureDestination point.
    """
    if not options:
        options = features.get("properties", {})

    coords = get_coords_from_features(features, ["Point"])

    bearing = options.get("bearing", 180)
    distance = options.get("dist", 100)
    units = options.get("units", "kilometers")

    distance_in_meters = convert_length(
        abs(distance), original_unit=units, final_unit="meters"
    )

    if distance < 0:
        distance_in_meters *= -1

    destination = calculate_rhumb_destination(coords, distance_in_meters, bearing)

    # compensate the crossing of the 180th meridian:
    #  (https://macwright.org/2016/09/26/the-180th-meridian.html)
    # solution from:
    # https://github.com/mapbox/mapbox-gl-js/issues/3250#issuecomment-294887678

    if (destination[0] - coords[0]) > 180:
        destination[0] -= 360
    elif (coords[0] - destination[0]) > 180:
        destination[0] += 360

    return point(destination, options.get("properties", None))


def calculate_rhumb_destination(
    origin: Sequence, distance_in_meters: float, bearing: float, radius: float = None
) -> List:
    """
    Calculates the destination point having travelled along a rhumb line
    from origin point the given distance on the  given bearing.
    Adapted from Geodesy:
    http://www.movable-type.co.uk/scripts/latlong.html#rhumblines

    param origin: point coordinates in [lng, lat] form
    param distance: - Distance travelled, in same units as earth radius (default: metres).
    param bearing: - Bearing in degrees from north.
    param radius: - (Mean) radius of earth

    returns destination: point.
    """
    if not radius:
        radius = earth_radius

    # angular distance in radians
    delta = distance_in_meters / radius
    # to radians, but without normalize to pi
    lambda_1 = origin[0] * pi / 180

    phi_1 = degrees_to_radians(origin[1])
    theta = degrees_to_radians(bearing)

    delta_phi = delta * cos(theta)
    phi_2 = phi_1 + delta_phi

    # check for some points going past the pole, normalise latitude if so
    if abs(phi_2) > (pi / 2) and (phi_2 > 0):
        phi_2 = pi - phi_2
    if abs(phi_2) > (pi / 2) and (phi_2 < 0):
        phi_2 = pi - phi_2

    delta_psi = log(tan(phi_2 / 2 + pi / 4) / tan(phi_1 / 2 + pi / 4))

    # E-W course becomes ill-conditioned with 0/0
    if abs(delta_psi) > 10e-12:
        q_1 = delta_phi / delta_psi
    else:
        q_1 = cos(phi_1)

    delta_lambda = delta * sin(theta) / q_1
    lambda_2 = lambda_1 + delta_lambda

    # normalise to −180..+180°
    destination = [
        fmod(((lambda_2 * 180 / pi) + 540), 360) - 180,
        (phi_2 * 180 / pi),
    ]

    return destination



from functools import reduce

from turf.invariant import get_coords_from_features
from turf.utils.helpers import get_input_dimensions


def bbox(features):
    """
    Takes a set of features and returns a bounding box containing of all input features.

    :param features: any GeoJSON feature or feature collection
    :return: bounding box extent in [minX, minY, maxX, maxY] order
    """

    bounding_box = [float("inf"), float("inf"), float("-inf"), float("-inf")]

    coords = get_coords_from_features(features)

    if get_input_dimensions(coords) == 1:
        coords = [coords]

    return reduce(reduce_coords, coords, bounding_box)


def reduce_coords(bounding_box, coord):
    input_dimension = get_input_dimensions(coord)

    if input_dimension >= 2:
        return reduce(lambda bb, c: reduce_coords(bb, c), coord, bounding_box)

    if bounding_box[0] > coord[0]:
        bounding_box[0] = coord[0]
    if bounding_box[1] > coord[1]:
        bounding_box[1] = coord[1]
    if bounding_box[2] < coord[0]:
        bounding_box[2] = coord[0]
    if bounding_box[3] < coord[1]:
        bounding_box[3] = coord[1]

    return bounding_box



from typing import Dict, Sequence, TypeVar, Union
from collections import deque

from rtree import index

from turf.helpers import (
    Feature,
    FeatureCollection,
    LineString,
    MultiLineString,
    MultiPolygon,
    Point,
    Polygon,
    get_input_dimensions,
)

from turf.envelope._envelope import envelope
from turf.polygon_to_line import polygon_to_line
from turf.helpers import feature, feature_collection, line_string, point
from turf.invariant import get_coords_from_features, get_geometry_type
from turf.utils.error_codes import error_code_messages
from turf.utils.exceptions import InvalidInput


LinePolyFeature = TypeVar(
    "LineFeature",
    Dict,
    Feature,
    FeatureCollection,
    LineString,
    MultiLineString,
    MultiPolygon,
    Polygon,
)


def line_intersect(input_1: LinePolyFeature, input_2: LinePolyFeature) -> Feature:
    """
    Takes any LineString or Polygon GeoJSON and returns the intersecting point(s).

    :param line_1: {GeoJSON} line1 any LineString or Polygon
    :param line_2:{GeoJSON} line2 any LineString or Polygon
    :returns: {FeatureCollection<Point>} point(s) that intersect both
    """
    line_1_segments = get_line_segments(input_1)
    line_2_segments = get_line_segments(input_2)

    possible_intersects = []
    intersects = []

    possible_intersects.extend(spatial_filtering(line_1_segments, line_2_segments))

    for i in possible_intersects:
        pnt = calculate_intersect(*i)

        if pnt:
            intersects.append(pnt)

    return feature_collection(intersects)


def spatial_filtering(line_1: Sequence, line_2: Sequence) -> Sequence:
    """
    Filters possible intersections of the lines via their bounding box

    :param line_1: line_1 coordinates of segments
    :param line_2: line_2 coordinates of segments
    :returns: list of line segments that possibly intersect with each other
    """
    possible_intersects = []
    rtree_index = index.Index()

    for i, seg in enumerate(line_1):
        bbox = envelope(seg)
        rtree_index.insert(i, bbox["bbox"])

    for j, seg in enumerate(line_2):
        bbox = envelope(seg)
        seg_intersection_idx = deque(rtree_index.intersection(bbox["bbox"]))

        while seg_intersection_idx:
            seg_idx = seg_intersection_idx.pop()

            possible_intersects.append([*line_1[seg_idx], *line_2[j]])

    return possible_intersects


def calculate_intersect(
    seg_1_start: Sequence,
    seg_1_end: Sequence,
    seg_2_start: Sequence,
    seg_2_end: Sequence,
) -> Union[None, Point]:
    """
    Calculates the intersection point of two segments otherwise None

    :param seg_1_start: coordinates of segment 1 start
    :param seg_1_end: coordinates of segment 1 end
    :param seg_2_start: coordinates of segment 2 start
    :param seg_2_end: coordinates of segment 2 end
    :returns: {Point} if intersection, otherwise None
    """
    x1 = seg_1_start[0]
    y1 = seg_1_start[1]
    x2 = seg_1_end[0]
    y2 = seg_1_end[1]
    x3 = seg_2_start[0]
    y3 = seg_2_start[1]
    x4 = seg_2_end[0]
    y4 = seg_2_end[1]

    pnt = None

    denom = ((y4 - y3) * (x2 - x1)) - ((x4 - x3) * (y2 - y1))
    numeA = ((x4 - x3) * (y1 - y3)) - ((y4 - y3) * (x1 - x3))
    numeB = ((x2 - x1) * (y1 - y3)) - ((y2 - y1) * (x1 - x3))

    if denom == 0:
        return pnt

    uA = numeA / denom
    uB = numeB / denom

    if uA >= 0 and uA <= 1 and uB >= 0 and uB <= 1:
        x = round(x1 + (uA * (x2 - x1)), 6)
        y = round(y1 + (uA * (y2 - y1)), 6)
        pnt = point([x, y])

    return pnt


def get_line_segments(line: LinePolyFeature) -> Sequence:
    """
    Gets segments from a line feature

    :param line: any LineString or Polygon
    :return: sequence of segmetns
    """
    segments = []

    geometry_type = get_geometry_type(line)

    if isinstance(geometry_type, str):
        geometry_type = [geometry_type]

    for line_geo in geometry_type:
        if line_geo in ["MultiPolygon", "Polygon"]:
            line = polygon_to_line(line)
            line_geo = line["geometry"]["type"]

        line_coords = get_coords_from_features(line, ("LineString", "MultiLineString"))

        if line_geo in ["LineString"] and get_input_dimensions(line_coords) == 2:
            line_coords = [line_coords]

        for line_coord in line_coords:
            segments.extend(list(zip(line_coord, line_coord[1:])))

    return segments



from typing import Dict, List, Sequence, TypeVar, Union
from copy import deepcopy

from turf.distance import distance
from turf.explode import explode
from turf.helpers import point, Point
from turf.helpers import Feature, FeatureCollection, Geometry
from turf.invariant import get_coords_from_features


GeoJson = TypeVar("GeoJson", Dict, Feature, FeatureCollection, Geometry)


def nearest_point(target: Union[Sequence, Dict, Feature], features: GeoJson) -> Point:
    """
    Calculates the closest reference point from a feature collection towards a target point
    This calculation is geodesic.

    :param target: targetPoint the reference point
    :param features: points against input point set
    :return: the closest point in the features set to the reference point
    """
    min_distance = float("inf")
    nearest_point = None
    feature_index = None

    features = explode(features)
    points = features.get("features")

    target = get_coords_from_features(target, ["Point"])

    for i, point in enumerate(points):
        dist = distance(target, point)
        if dist < min_distance:
            min_distance = dist
            nearest_point = deepcopy(point)
            feature_index = i

    if "properties" in nearest_point:
        nearest_point["properties"].update(
            {"featureIndex": feature_index, "distanceToPoint": min_distance}
        )
    else:
        nearest_point.update(
            {
                "properties": {
                    "featureIndex": feature_index,
                    "distanceToPoint": min_distance,
                }
            }
        )

    return nearest_point



from typing import List, TypeVar, Union

from turf.boolean_point_in_polygon import boolean_point_in_polygon
from turf.center import center
from turf.explode import explode
from turf.invariant import get_coords_from_features, get_geometry_type

from turf.nearest_point import nearest_point
from turf.helpers import Feature, FeatureCollection, Point
from turf.helpers import feature, feature_collection, point, polygon, multi_polygon
from turf.helpers._features import all_geometry_types
from turf.utils.error_codes import error_code_messages
from turf.utils.exceptions import InvalidInput

from turf.boolean_point_on_line import boolean_point_on_line

GeoJSON = TypeVar("GeoJSON", Feature, FeatureCollection)


def point_on_feature(features: GeoJSON) -> Point:
    """
    Takes a Feature or FeatureCollection and returns a {Point} guaranteed to be on the surface of the feature.

    Given a {Polygon}, the point will be in the area of the polygon
    Given a {LineString}, the point will be along the string
    Given a {Point}, the point will the same as the input

    :param features: any GeoJSON feature or feature collection
    :return: Point GeoJSON Feature on the surface of `input`
    """
    feature_collection = normalize_to_feature_collection(features)

    center_point = center(feature_collection)
    center_coords = center_point.get("geometry").get("coordinates")

    # check to see if centroid is on surface
    center_on_surface = False

    geometry_type = get_geometry_type(feature_collection)
    geometry_coords = get_coords_from_features(feature_collection)

    if isinstance(geometry_type, str):
        geometry_type = [geometry_type]

    for geo_type, geo_coords in zip(geometry_type, geometry_coords):
        if geo_type in ["Point", "MultiPoint"]:
            if geo_type == "Point":
                geo_coords = [geo_coords]

            for point_coords in geo_coords:
                if (center_coords[0] == point_coords[0]) and (
                    center_coords[1] == point_coords[1]
                ):
                    center_on_surface = True
                    break

        elif geo_type in ["LineString", "MultiLineString"]:
            if geo_type == "LineString":
                geo_coords = [geo_coords]

            for line_coords in geo_coords:
                if boolean_point_on_line(center_coords, line_coords):
                    center_on_surface = True
                    break

        elif geo_type in ["Polygon", "MultiPolygon"]:
            if geo_type == "Polygon":
                geo_coords = polygon(geo_coords)
            else:
                geo_coords = multi_polygon(geo_coords)
            if boolean_point_in_polygon(center_point, geo_coords):
                center_on_surface = True
                break

    if center_on_surface:
        point_on_surface = center_point

    else:
        point_on_surface = nearest_point(center_point, feature_collection)

    return point_on_surface


def normalize_to_feature_collection(geojson: GeoJSON) -> FeatureCollection:
    """
    Normalizes any GeoJSON to a FeatureCollection

    :param geojson: any GeoJSON
    :return: FeatureCollection
    """
    geojson_type = geojson.get("type")

    if geojson_type == "FeatureCollection":
        pass

    elif geojson_type == "Feature":
        geojson = feature_collection([geojson])

    else:
        geojson = feature_collection([feature(geojson)])

    return geojson



from turf.helpers import polygon
from turf.utils.error_codes import error_code_messages
from turf.utils.exceptions import InvalidInput


def bbox_polygon(bbox, options=None):
    """
    Takes a bounding box and returns an equivalent Polygon feature.

    :param bbox: bounding box extent in [minX, minY, maxX, maxY] order
    :param options: optional parameters
        [options["properties"]={}] Translate GeoJSON Properties to Point
        [options["id"]={}] Translate GeoJSON Id to Point
    :return: a Polygon representation of the bounding box
    """

    if not options:
        options = {}

    if not isinstance(bbox, list) or len(bbox) != 4:
        raise InvalidInput(error_code_messages["InvalidBoundingBox"])

    west = float(bbox[0])
    south = float(bbox[1])
    east = float(bbox[2])
    north = float(bbox[3])

    low_left = [west, south]
    top_left = [west, north]
    top_right = [east, north]
    low_right = [east, south]

    return polygon(
        [
            [
                low_left,
                low_right,
                top_right,
                top_left,
                low_left,
            ]
        ],
        options.get("properties", None),
        {"bbox": bbox, "id": options.get("id", None)},
    )



from typing import Dict, List, Sequence, Union

from turf.distance import distance
from turf.rhumb_distance import rhumb_distance
from turf.helpers import convert_length, dot, Feature
from turf.invariant import get_coords_from_features


def point_to_line_distance(
    point: Union[Sequence, Dict, Feature],
    line: Union[Sequence, Dict, Feature],
    options: Dict = None,
) -> float:
    """
    Returns the minimum distance between a {Point} and a {LineString},
    being the distance from a line the minimum distance between the point and
    any segment of the `LineString`

    http://geomalgorithms.com/a02-_lines.html

    :param point: Point GeoJSON Feature or Geometry
    :param line: LineString GeoJSON Feature or Geometry
    :param options: Optional parameters
        [options["units"]]: any supported unit (e.g. degrees, radians, miles...)
        [options["method"]]: geodesic or 'planar for distance calculation

    :return: distance between point and line
    """
    dist = []
    if not isinstance(options, dict):
        options = {"method": "geodesic", "units": "kilometers"}

    point = get_coords_from_features(point, ["Point"])
    line = get_coords_from_features(line, ["LineString"])

    for i in range(1, len(line)):
        dist.append(
            get_distance_to_segment(
                point, line[i - 1], line[i], options.get("method", "geodesic")
            )
        )

    dist = convert_length(min(dist), "degrees", options.get("units", "kilometers"))

    return dist


def get_distance_to_segment(
    point: List, segment_start: List, segment_end: List, method: str
):
    """
    Calculates the distance from the point to the segment.

    :param point: Point coordinates
    :param segment_start: segment start point from the line feature
    :param segment_end: adjacent segment end point from the line feature
    :param method: wether to calculate the distance based on geodesic
                    (spheroid) or planar (flat) method.
                          Valid options are: 'geodesic' or 'planar

    :return: distance between point and both segments"""
    v = [segment_end[0] - segment_start[0], segment_end[1] - segment_start[1]]

    w = [point[0] - segment_start[0], point[1] - segment_start[1]]

    c1 = dot(w, v)
    c2 = dot(v, v)

    b2 = c1 / c2
    point_2 = [segment_start[0] + (b2 * v[0]), segment_start[1] + (b2 * v[1])]

    if c1 <= 0:
        dist = calculate_distance_by_method(point, segment_start, method)
    elif c2 <= c1:
        dist = calculate_distance_by_method(point, segment_end, method)
    else:
        dist = calculate_distance_by_method(point, point_2, method)

    return dist


def calculate_distance_by_method(point_1: List, point_2: List, method: str) -> float:
    """
    Wrapper to calculate the distance between two points. Depending on the
    method, if calls rhumb distance or haversine distance

    :param point_1: Point coordinates
    :param point_2: Point coordinates
    :param method: wether to calculate the distance based on geodesic
                          (spheroid) or planar (flat) method.
                          Valid options are: 'geodesic' or 'planar

    :return: distance between point and both segments in radians
    """

    if method == "planar":
        dist = rhumb_distance(point_1, point_2, {"units": "degrees"})
    else:
        dist = distance(point_1, point_2, {"units": "degrees"})

    return dist



from math import pi
from typing import List

from turf.helpers._units import factors, area_factors
from turf.utils.exceptions import InvalidInput
from turf.utils.error_codes import error_code_messages


def c_like_modulo(number, base):
    return number - int(number / base) * base


def dot(vector_1: List, vector_2: List) -> float:
    """
    Dot product of two lists

    :param vector_1: list of values
    :param vector_2: list of values

    :return: dot product
    """
    if len(vector_1) != len(vector_2):
        raise InvalidInput(error_code_messages["InvalidLength"])

    return sum(x * y for x, y in zip(vector_1, vector_2))


def degrees_to_radians(degrees):
    """
    :param degrees: degrees angle
    :return: angle in radians
    """

    if not isinstance(degrees, (float, int)):
        raise InvalidInput(error_code_messages["InvalidDegrees"])

    radians = c_like_modulo(degrees, 360)
    return radians * pi / 180


def radians_to_degrees(radians):
    """
    :param radians: radians angle in radians
    :return: degrees between 0 and 360
    """

    if not isinstance(radians, (float, int)):
        raise InvalidInput(error_code_messages["InvalidRadians"])

    degrees = c_like_modulo(radians, 2 * pi)
    return degrees * 180 / pi


def length_to_radians(distance, units="kilometers"):
    """
    :param distance: distance in real units
    :param units: units of the distance. Can be degrees, radians, miles, kilometers,
            inches, yards, metres, meters, kilometres, kilometers. Defaults to kilometers
    :return: radians
    """

    if not isinstance(distance, (float, int)):
        raise (InvalidInput(error_code_messages["InvalidDistance"]))

    try:
        factor = factors[units]
    except KeyError:
        raise InvalidInput(error_code_messages["InvalidUnits"](units))

    return distance / factor


def radians_to_length(radians, units="kilometers"):
    """
    :param radians: radians in radians across the sphere
    :param units: units of the distance. Can be degrees, radians, miles, kilometers,
            inches, yards, metres, meters, kilometres, kilometers. Defaults to kilometers
    :return: distance
    """

    try:
        factor = factors[units]
    except KeyError:
        raise InvalidInput(error_code_messages["InvalidUnits"](units))

    return radians * factor


def length_to_degrees(distance, units="kilometers"):
    """
    :param distance: distance in real units
    :param units: units of the distance. Can be degrees, radians, miles, kilometers,
            inches, yards, metres, meters, kilometres, kilometers. Defaults to kilometers
    :return: degrees
    """

    return radians_to_degrees(length_to_radians(distance, units))


def convert_length(length, original_unit="kilometers", final_unit="kilometers"):
    """
    :param length: length to be converted
    :param original_unit: original unit of the length
    :param final_unit: return unit of the length
    :return: the converted length
    """

    if not isinstance(length, (float, int)) or length < 0:
        raise InvalidInput(error_code_messages["InvalidLength"])

    return radians_to_length(length_to_radians(length, original_unit), final_unit)


def convert_area(area, original_unit="meters", final_unit="kilometers"):
    """
    :param area: area to be converted
    :param original_unit: original unit of the area
    :param final_unit: returned unit of area
    :return: the converted area
    """

    if not isinstance(area, (float, int)) or area < 0:
        raise InvalidInput(error_code_messages["InvalidArea"])

    try:
        start_factor = area_factors[original_unit]
    except KeyError:
        raise InvalidInput(
            InvalidInput(error_code_messages["InvalidUnits"](original_unit))
        )

    try:
        final_factor = area_factors[final_unit]
    except KeyError:
        raise InvalidInput(
            InvalidInput(error_code_messages["InvalidUnits"](final_unit))
        )

    return (area / start_factor) * final_factor



from turf.helpers import point
from turf.bearing import bearing
from turf.destination import destination
from turf.distance import distance
from turf.invariant import get_coords_from_geometry, get_coords_from_features
from turf.utils.exceptions import InvalidInput
from turf.utils.error_codes import error_code_messages
from turf.utils.helpers import truncate


def along(line, dist, options=None):
    """
    Takes a LineString and returns a Point at a specified distance along the line

    :param line: input LineString
    :param dist: distance along the line
    :param options: optional parameters
        [options["units"]="kilometers"] can be degrees, radians, miles, or kilometers
    :return: Point `dist` `units` along the line
    """

    if not options or not isinstance(options, dict):
        options = {}

    if not isinstance(dist, (float, int)) or dist < 0:
        raise InvalidInput(error_code_messages["InvalidDistance"])

    coords = get_coords_from_features(line, ["LineString"])

    travelled = 0
    for i in range(len(coords)):
        if dist >= travelled and i == len(coords) - 1:
            break
        elif travelled >= dist:
            overshot = dist - travelled
            if not overshot:
                return point([truncate(coord, 6) for coord in coords[i]])
            else:
                direction = bearing(coords[i], coords[i - 1]) - 180
                interpolated = destination(coords[i], overshot, direction, options)
                return interpolated
        else:
            travelled += distance(coords[i], coords[i + 1])

    return point([truncate(coord, 6) for coord in coords[-1]])



from functools import reduce

from turf.helpers import point
from turf.invariant import get_coords_from_features
from turf.utils.helpers import get_input_dimensions


def centroid(features, options=None):
    """
    Takes one or more features and calculates the centroid using the mean of all vertices.
    This lessens the effect of small islands and artifacts when calculating the centroid of a set of polygons.

    :param features: GeoJSON features to be centered
    :param options: optional parameters
        [options["properties"]={}] Translate GeoJSON Properties to Point
    :return: a Point feature corresponding to the centroid of the input features
    """

    if not options:
        options = {}

    coords = get_coords_from_features(features)

    if get_input_dimensions(coords) == 1:
        coords = [coords]

    x_sum = 0
    y_sum = 0
    length = 0

    x_sum, y_sum, length = reduce(reduce_coords, coords, [x_sum, y_sum, length])

    return point([x_sum / length, y_sum / length], options.get("properties", None))


def reduce_coords(sum_array, coords):
    input_dimension = get_input_dimensions(coords)

    if input_dimension >= 2:
        return reduce(lambda prev, coord: reduce_coords(prev, coord), coords, sum_array)

    return [sum_array[0] + coords[0], sum_array[1] + coords[1], sum_array[2] + 1]



from math import atan2, fmod, log, pi, tan
from typing import Dict, Sequence, Union

from turf.helpers import degrees_to_radians, radians_to_degrees
from turf.helpers import Feature
from turf.invariant import get_coords_from_features


def rhumb_bearing(
    origin: Union[Sequence, Dict, Feature],
    destination: Union[Sequence, Dict, Feature],
    options: Dict = None,
) -> float:
    """
    Takes two {Point|points} and finds the bearing angle between them along a
    Rhumb line
    * i.e. the angle measured in degrees start the north line (0 degrees)

    https://en.wikipedia.org/wiki/Rhumb_line

    :param start: starting point [lng, lat] or Point feature
    :param end: ending point [lng, lat] or Point feature
    :param options: Optional parameters
        [options["final"]]: Calculates the final bearing if True

    :return: bearing from north in decimal degrees
    """
    if not isinstance(options, dict):
        options = {}

    origin = get_coords_from_features(origin, ["Point"])
    destination = get_coords_from_features(destination, ["Point"])
    final = options.get("final", False)

    if final:
        bearing = calculate_rhumb_bearing(destination, origin)
    else:
        bearing = calculate_rhumb_bearing(origin, destination)

    return bearing


def calculate_rhumb_bearing(origin: Sequence, destination: Sequence) -> float:
    """
    Calculates the bearing from origin to destination point along a rhumb line.
    http://www.edwilliams.org/avform.htm#Rhumb
    """
    phi_1 = degrees_to_radians(origin[1])
    phi_2 = degrees_to_radians(destination[1])
    delta_lambda = degrees_to_radians(destination[0] - origin[0])

    # if delta_lambda over 180° take shorter rhumb line across the anti-meridian:
    if abs(delta_lambda) > pi:
        if delta_lambda > 0:
            delta_lambda = -(2 * pi - delta_lambda)
        if delta_lambda < 0:
            delta_lambda = 2 * pi + delta_lambda

    delta_psi = log(tan(phi_2 / 2 + pi / 4) / tan(phi_1 / 2 + pi / 4))
    theta = atan2(delta_lambda, delta_psi)

    return fmod(radians_to_degrees(theta) + 360, 360)



from turf.bearing import bearing
from turf.destination import destination
from turf.distance import distance


def midpoint(point1, point2):
    """
    Takes two point features and returns a point midway between them.
    The midpoint is calculated geodesically, meaning the curvature of the earth is taken into account.

    :param point1: first point
    :param point2: second point
    :return: a point midway between point 1 and point 2
    """

    dist = distance(point1, point2)
    heading = bearing(point1, point2)
    mid_point = destination(point1, dist / 2, heading)

    return mid_point



from typing import Dict, Sequence, TypeVar

from turf.helpers import (
    Feature,
    FeatureCollection,
    LineString,
    MultiLineString,
    MultiPolygon,
    Polygon,
)

from turf.helpers import feature_collection, line_string, multi_line_string
from turf.invariant import get_coords_from_features, get_geometry_type
from turf.utils.error_codes import error_code_messages
from turf.utils.exceptions import InvalidInput


PolygonFeature = TypeVar("PolygonFeature", Dict, Polygon, MultiPolygon)
LineFeature = TypeVar(
    "LineFeature", Dict, Feature, FeatureCollection, LineString, MultiLineString
)


def polygon_to_line(polygon: PolygonFeature, options: Dict = {}) -> LineFeature:
    """Converts a {Polygon} to a {LineString} or a {MultiPolygon} to a {FeatureCollection}
    or {MultiLineString}.

    :param polygon: Feature to convert
    :param options: Optional parameters

    :return:
        {Feature Collection|LineString|MultiLineString} of converted (Multi)Polygon to (Multi)LineString
    """
    if not options:
        properties = polygon.get("properties", {})
    else:
        properties = options.get("properties", {})

    geometry_type = get_geometry_type(polygon, ("Polygon", "MultiPolygon"))

    polygon_coords = get_coords_from_features(polygon, ("Polygon", "MultiPolygon"))

    if isinstance(geometry_type, str):
        geometry_type = [geometry_type]
        polygon_coords = [polygon_coords]

    for geo_type, poly_coords in zip(geometry_type, polygon_coords):
        if geo_type == "MultiPolygon":
            line_coords = []

            for poly_coord in poly_coords:
                line_coords.append(coords_to_line(poly_coord, properties))

            line_feature = feature_collection(line_coords)

        else:
            line_feature = coords_to_line(poly_coords, properties)

    return line_feature


def coords_to_line(coords: Sequence, properties: Dict) -> LineFeature:
    """Converts coordinates to a {LineString}} or {MultiLineString}.

    :param polygon: Feature to convert
    :param properties: Optional parameters

    :return:
        {feature} of a (Multi)LineString
    """

    if len(coords) > 1:
        feature = multi_line_string(coords, properties)

    else:
        feature = line_string(coords[0], properties)

    return feature



from typing import Dict, List, Sequence, TypeVar

from turf.invariant import get_coords_from_features, get_coords_from_geometry

from turf.helpers import feature_collection, point
from turf.helpers import Feature, FeatureCollection, Geometry

from turf.utils.helpers import get_input_dimensions
from turf.utils.error_codes import error_code_messages
from turf.utils.exceptions import InvalidInput
from turf.helpers._features import all_geometry_types

GeoJson = TypeVar("GeoJson", Dict, Feature, FeatureCollection, Geometry)


def explode(features: GeoJson) -> FeatureCollection:
    """
    Takes a feature or set of features and returns all positions as {Point|points}.

    :param features: any GeoJSON feature or feature collection
    :return: {FeatureCollection} points representing the exploded input features
    """
    points = []

    try:
        geojson_type = features.get("type")
    except AttributeError:
        raise InvalidInput(error_code_messages["InvalidGeometry"](all_geometry_types))

    if geojson_type in ["FeatureCollection", "GeometryCollection"]:
        key = "features" if geojson_type == "FeatureCollection" else "geometries"

        for feature in features[key]:
            properties = feature.get("properties", {})
            coords = get_coords_from_features(feature)
            points.extend(reduce_coordinates_to_points(coords, properties))

    else:
        properties = features.get("properties", {})
        coords = get_coords_from_geometry(features)
        points.extend(reduce_coordinates_to_points(coords, properties))

    return feature_collection(points)


def reduce_coordinates_to_points(
    coords: Sequence, properties: Dict
) -> Sequence[Feature]:
    """
    Transforms dimensionality of the incoming coordinates into Point Features

    :param coords: any sequence of coordinates
    :param properties: properties of the coordinates sequence
    :return: {Feature} points of transformed coordinates
    """

    dim = get_input_dimensions(coords)

    if dim == 1:
        coords = [coords]

    while dim > 2:
        coords = [c for coord in coords for c in coord]
        dim = get_input_dimensions(coords)

    points = [point(coord, properties) for coord in coords]

    return points



from typing import Sequence, Union, Dict

from turf.helpers import Feature
from turf.invariant import (
    get_coords_from_features,
    get_geometry_from_features,
    get_geometry_type,
)
from turf.bbox import bbox as bounding_box
from turf.utils.error_codes import error_code_messages
from turf.utils.exceptions import InvalidInput

valid_polygons = ["Polygon", "MultiPolygon"]


def boolean_point_in_polygon(
    point: Union[Sequence, Dict, Feature],
    polygon: Union[Dict, Feature],
    options: Dict = None,
):
    """
    Takes a {@link Point} and a Polygon or MultiPolygon and determines if the point
    resides inside the polygon. The polygon can be convex or concave. The function accounts for holes.

    reference:

    http://en.wikipedia.org/wiki/Even%E2%80%93odd_rule
    modified from: https://github.com/substack/point-in-polygon/blob/master/index.js
    which was modified from http://www.ecse.rpi.edu/Homepages/wrf/Research/Short_Notes/pnpoly.html

    :param point: input Point Feature
    :param polygon: input Polygon or MultiPolygon Feature
    :param options: optional parameters
        [options["ignoreBoundary"]] True if polygon boundary should be ignored when determining if
                                    the point is inside the polygon otherwise False.
    :return: True if the Point is inside the Polygon; False otherwise
    """
    if not isinstance(options, dict):
        options = {}

    ignore_boundary = options.get("ignoreBoundary", False)

    point_coords = get_coords_from_features(point, ["Point"])
    polygon_coords = get_coords_from_features(polygon, valid_polygons)

    geometry_type = get_geometry_type(polygon, valid_polygons)

    bbox = bounding_box(polygon)

    if not in_bbox(point_coords, bbox):
        return False

    if isinstance(geometry_type, str):
        geometry_type = [geometry_type]
        polygon_coords = [polygon_coords]

    inside_polygon = False

    for geo_type, poly_coords in zip(geometry_type, polygon_coords):
        if geo_type == "Polygon":
            poly_coords = [poly_coords]

        for polygon in poly_coords:
            # check if it is in the outer ring first
            if in_ring(point_coords, polygon[0], ignore_boundary):
                in_hole = False

                for ring in polygon[1:]:
                    if in_ring(point_coords, ring, not ignore_boundary):
                        in_hole = True

                if not in_hole:
                    inside_polygon = True

            if inside_polygon:
                break

    return inside_polygon


def in_ring(point: Sequence, ring: Sequence, ignore_boundary: bool):
    """
    Checks if point is inside a ring
    :param point: point coordinates [x, y]
    :param ring: ring [[x, y], [x, y], ...]
    :param ignore_boundary: True if polygon boundary should be ignored when determining if
    the point is inside the polygon otherwise False.
    :return: True if point is inside, False otherwise
    """

    is_inside = False

    if ring[0][0] == ring[-1][0] and ring[0][1] == ring[-1][1]:
        ring = ring[:-1]

    i = 0
    j = len(ring) - 1
    while i < len(ring):
        xi = ring[i][0]
        yi = ring[i][1]
        xj = ring[j][0]
        yj = ring[j][1]

        on_boundary = (
            point[1] * (xi - xj) + yi * (xj - point[0]) + yj * (point[0] - xi) == 0
            and ((xi - point[0]) * (xj - point[0]) <= 0)
            and ((yi - point[1]) * (yj - point[1]) <= 0)
        )

        if on_boundary:
            return not ignore_boundary

        intersect = ((yi > point[1]) != (yj > point[1])) and (
            point[0] < (xj - xi) * (point[1] - yi) / (yj - yi) + xi
        )

        if intersect:
            is_inside = not is_inside

        j = i
        i += 1

    return is_inside


def in_bbox(point: Sequence, bbox: Sequence):
    """
    Checks if point is inside bbox

    :param point: point coordinates [lng, lat]
    :param bbox: bbox [west, south, east, north]
    :return: True if point is inside, False otherwise
    """
    return bbox[0] <= point[0] <= bbox[2] and bbox[1] <= point[1] <= bbox[3]



from typing import Dict, List, Union

from turf.distance import distance
from turf.bbox import bbox
from turf.boolean_intersects import boolean_intersects
from turf.helpers import feature_collection, polygon, FeatureCollection


def triangle_grid(
    bbox: List[float],
    cell_side: Union[int, float],
    options: Dict = {},
) -> FeatureCollection:
    """
    Creates a square of rectangles from a bounding box, Feature or FeatureCollection.

    :param bbox: Array extent in [minX, minY, maxX, maxY] order
    :param cell_side: dimension of each cell
    :param options: Optional parameters
        [options["units"]]: units ("degrees", "radians", "miles", "kilometers")
                            of the given cell_width and cell_height
        [options["mask"]]: if passed a Polygon or MultiPolygon here,
                           the grid Points will be created only inside it
        [options["properties"]]: passed to each point of the grid

    :returns: FeatureCollection of a grid of polygons
    """
    if not isinstance(options, dict):
        options = {}

    results = []
    west = bbox[0]
    south = bbox[1]
    east = bbox[2]
    north = bbox[3]

    x_fraction = cell_side / (distance([west, south], [east, south], options))
    cell_width_deg = x_fraction * (east - west)
    y_fraction = cell_side / (distance([west, south], [west, north], options))
    cell_height_deg = y_fraction * (north - south)

    # if the grid does not fill the bbox perfectly, center it.
    xi = 0
    current_x = west
    while current_x <= east:
        yi = 0
        current_y = south
        while current_y <= north:
            cell_triangle1 = None
            cell_triangle2 = None

            if (xi % 2 == 0) and (yi % 2 == 0):
                cell_triangle1 = polygon(
                    [
                        [
                            [current_x, current_y],
                            [current_x, current_y + cell_height_deg],
                            [current_x + cell_width_deg, current_y],
                            [current_x, current_y],
                        ]
                    ],
                    options.get("properties", {}),
                )

                cell_triangle2 = polygon(
                    [
                        [
                            [current_x, current_y + cell_height_deg],
                            [current_x + cell_width_deg, current_y + cell_height_deg],
                            [current_x + cell_width_deg, current_y],
                            [current_x, current_y + cell_height_deg],
                        ]
                    ],
                    options.get("properties", {}),
                )

            elif (xi % 2 == 0) and (yi % 2 == 1):
                cell_triangle1 = polygon(
                    [
                        [
                            [current_x, current_y],
                            [current_x + cell_width_deg, current_y + cell_height_deg],
                            [current_x + cell_width_deg, current_y],
                            [current_x, current_y],
                        ]
                    ],
                    options.get("properties", {}),
                )

                cell_triangle2 = polygon(
                    [
                        [
                            [current_x, current_y],
                            [current_x, current_y + cell_height_deg],
                            [current_x + cell_width_deg, current_y + cell_height_deg],
                            [current_x, current_y],
                        ]
                    ],
                    options.get("properties", {}),
                )

            elif (yi % 2 == 0) and (xi % 2 == 1):
                cell_triangle1 = polygon(
                    [
                        [
                            [current_x, current_y],
                            [current_x, current_y + cell_height_deg],
                            [current_x + cell_width_deg, current_y + cell_height_deg],
                            [current_x, current_y],
                        ]
                    ],
                    options.get("properties", {}),
                )

                cell_triangle2 = polygon(
                    [
                        [
                            [current_x, current_y],
                            [current_x + cell_width_deg, current_y + cell_height_deg],
                            [current_x + cell_width_deg, current_y],
                            [current_x, current_y],
                        ]
                    ],
                    options.get("properties", {}),
                )

            elif (yi % 2 == 1) and (xi % 2 == 1):
                cell_triangle1 = polygon(
                    [
                        [
                            [current_x, current_y],
                            [current_x, current_y + cell_height_deg],
                            [current_x + cell_width_deg, current_y],
                            [current_x, current_y],
                        ]
                    ],
                    options.get("properties", {}),
                )

                cell_triangle2 = polygon(
                    [
                        [
                            [current_x, current_y + cell_height_deg],
                            [current_x + cell_width_deg, current_y + cell_height_deg],
                            [current_x + cell_width_deg, current_y],
                            [current_x, current_y + cell_height_deg],
                        ]
                    ],
                    options.get("properties", {}),
                )

            if "mask" in options:
                if boolean_intersects(options["mask"], cell_triangle1):
                    results.append(cell_triangle1)
                if boolean_intersects(options["mask"], cell_triangle2):
                    results.append(cell_triangle2)
            else:
                results.append(cell_triangle1)
                results.append(cell_triangle2)

            current_y += cell_height_deg
            yi += 1

        current_x += cell_width_deg
        xi += 1

    return feature_collection(results)



from turf.along import along
from turf.area import area
from turf.bbox import bbox
from turf.bbox_polygon import bbox_polygon
from turf.bearing import bearing
from turf.boolean_disjoint import boolean_disjoint
from turf.boolean_intersects import boolean_intersects
from turf.boolean_point_in_polygon import boolean_point_in_polygon
from turf.boolean_point_on_line import boolean_point_on_line
from turf.boolean_within import boolean_within
from turf.center import center
from turf.centroid import centroid
from turf.destination import destination
from turf.distance import distance
from turf.envelope import envelope
from turf.explode import explode
from turf.great_circle import great_circle
from turf.helpers import *
from turf.hex_grid import hex_grid
from turf.length import length
from turf.line_intersect import line_intersect
from turf.midpoint import midpoint
from turf.nearest_point import nearest_point
from turf.point_grid import point_grid
from turf.point_on_feature import point_on_feature
from turf.point_to_line_distance import point_to_line_distance
from turf.polygon_tangents import polygon_tangents
from turf.polygon_to_line import polygon_to_line
from turf.rectangle_grid import rectangle_grid
from turf.rhumb_bearing import rhumb_bearing
from turf.rhumb_destination import rhumb_destination
from turf.rhumb_distance import rhumb_distance
from turf.square import square
from turf.square_grid import square_grid
from turf.triangle_grid import triangle_grid
from turf.version import __version__
